﻿using Microsoft.AspNetCore.Mvc;
using Newtonsoft.Json;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using Wicture.DbRESTFul.Infrastructure.Repository;
using Wicture.DbRESTFul.Resources.Api;
using Wicture.DbRESTFul.Resources.Csi;

namespace Wicture.DbRESTFul.PMT
{
    [Route("api")]
    public class ImportController : Controller
    {
        private readonly DbRESTFulRepository _repository;

        public ImportController(DbRESTFulRepository repository)
        {
            _repository = repository;
        }

        [HttpPost("ca/import")]
        public async Task<object> ImportApi(int projectId)
        {
            if (projectId <= 0)
            {
                return new StandardResult(ResultCode.LogicError, "No projectId specified.");
            }

            if (Request.Form.Files.Count == 0)
            {
                return new StandardResult(ResultCode.LogicError, "No file uploaded.");
            }

            int total = 0;
            int replace = 0;

            IEnumerable<dynamic> modules = await _repository.InvokeAsync("ListModulesForProject", new { projectId });

            foreach (var file in Request.Form.Files)
            {
                using (StreamReader reader = new StreamReader(file.OpenReadStream()))
                {
                    var text = reader.ReadToEnd();
                    var apis = JsonConvert.DeserializeObject<List<ApiModel>>(text);

                    foreach (var api in apis)
                    {
                        var checkResult = await _repository.InvokeAsync("CheckNameExsits", new { projectId, api.name }) as IEnumerable<dynamic>;
                        if (checkResult.Any())
                        {
                            replace++;
                            await _repository.InvokeAsync("DeleteApi", new { ids = checkResult.Select(d => d.id).OfType<int>().ToList() });

                            // TODO: 改成提示用户是否要覆盖。
                            //return AjaxResult.SetError($"API:[{api.name}] is already exsiting.", ResultCode.LogicError);
                        }
                    }

                    await _repository.InvokeAsync("ImportApis", apis.Select(ca => BeParameterized(projectId, ca, modules)));
                    total += apis.Count;
                }
            }

            return new StandardResult { Data = new { total, replace } };
        }

        private object BeParameterized(int projectId, ApiModel api, IEnumerable<dynamic> modules)
        {
            int? moduleId = modules == null ? null : modules.FirstOrDefault(a => a.name == api.module)?.id;

            return new
            {
                projectId,
                api.version,
                api.owner,
                api.createdAt,
                api.name,
                api.module,
                moduleId,
                api.url,
                useAbsoluteUrl = api.useAbsoluteUrl ? 1 : 0,
                api.method,
                api.title,
                api.summary,
                api.note,
                allowAnonymous = api.allowAnonymous ? 1 : 0,
                cache = JsonConvert.SerializeObject(api.cache),

                implemented = api.implemented ? 1 : 0,
                implementation = JsonConvert.SerializeObject(api.implementation),
                parameter = JsonConvert.SerializeObject(api.parameter),
                result = JsonConvert.SerializeObject(api.result),
                mock = JsonConvert.SerializeObject(api.mock)
            };
        }

        [HttpPost("csi/import")]
        public async Task<object> ImportCSI(int projectId)
        {
            if (projectId <= 0)
            {
                return new StandardResult(ResultCode.LogicError, "No projectId specified.");
            }

            if (Request.Form.Files.Count == 0)
            {
                return new StandardResult(ResultCode.LogicError, "No file uploaded.");
            }

            int total = 0;
            int replace = 0;

            IEnumerable<dynamic> modules = await _repository.InvokeAsync("ListModulesForProject", new { projectId });

            foreach (var file in Request.Form.Files)
            {
                using (StreamReader reader = new StreamReader(file.OpenReadStream()))
                {
                    var text = reader.ReadToEnd();
                    var csiList = JsonConvert.DeserializeObject<List<CsiModel>>(text);

                    foreach (var api in csiList)
                    {
                        var checkResult = await _repository.InvokeAsync("CheckCSINameExsits", new { projectId, api.name }) as IEnumerable<dynamic>;
                        if (checkResult.Any())
                        {
                            replace++;
                            await _repository.InvokeAsync("DeleteCSI", new { ids = checkResult.Select(d => d.id).OfType<int>().ToList() });
                        }
                    }

                    await _repository.InvokeAsync("ImportCSIs", csiList.Select(csi => BeParameterized(projectId, csi, modules)));
                    total += csiList.Count;
                }
            }

            return new StandardResult { Data = new { total, replace } };
        }

        private object BeParameterized(int projectId, CsiModel csi, IEnumerable<dynamic> modules)
        {
            int? moduleId = modules == null ? null : modules.FirstOrDefault(a => a.name == csi.module)?.id;

            return new
            {
                projectId,
                csi.name,
                moduleId,
                code = JsonConvert.SerializeObject(csi.code),
                rawCode = JsonConvert.SerializeObject(csi.code),
                csi.resultSet,
                csi.queryOnly,
                csi.requiredTransaction,
                middleWares = JsonConvert.SerializeObject(csi.middlewares),
                updatedBy = _repository.Context.Identity.Id,
                createdBy = _repository.Context.Identity.Id
            };
        }
    }

}
