﻿using Microsoft.AspNetCore.Hosting;
using Microsoft.AspNetCore.Mvc;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using Wicture.DbRESTFul.Infrastructure.Repository;
using Wicture.DbRESTFul.PMT.SpecGenerators;

namespace Wicture.DbRESTFul.PMT.Controllers
{
    public class SpecController : Controller
    {
        private readonly string docRoot = string.Empty;
        private readonly DbRESTFulRepository _repository;

        public SpecController(IWebHostEnvironment env, DbRESTFulRepository repository)
        {
            _repository = repository;
            docRoot = Path.Combine(env.ContentRootPath, @"doc/");
        }

        [HttpGet]
        [Route("api/spec/tables")]
        public object LoadTablesForProject(string connectionString,int projectId)
        {
            var helper = new SpecHelper(connectionString, projectId);
            var file = helper.ListTables().Select(t => new { name = t.name, title = t.title });
            return new StandardResult { Data = file };
        }

        /// <summary>
        /// 生成数据库文档
        /// </summary>
        /// <param name="config">
        /// {
        ///     "moduleName": "文档的模块名称",
        ///     "projectName": "项目名称",
        ///     "tables": [
        ///         {   
        ///             "name": "",
        ///             "title": ""
        ///         }
        ///     ...]
        /// }
        /// </param>
        /// <returns></returns>
        [HttpPost]
        [Route("api/spec/db")]
        public async Task<object> GenerateDbSpec()
        {
            using var reader = new StreamReader(Request.Body);
            var body = await reader.ReadToEndAsync();
            JObject config = JObject.Parse(body);

            var generator = new DbSpecGenerator(config);
            var file = generator.Generate(docRoot);
            return new StandardResult { Data = file };
        }

        /// <summary>
        /// 生成项目简介文档
        /// </summary>
        /// <param name="id"></param>
        /// <returns></returns>
        [HttpPost]
        [Route("api/spec/project/introduction")]
        public async Task<object> GenerateProjectIntroducation(int id)
        {
            var project = await _repository.InvokeAsync("GetProject", new { id });
            var generator = new IntroductionGenerator(project.name, project.introduction);
            var file = generator.Generate(docRoot);
            return new StandardResult { Data = file };
        }

        /// <summary>
        /// 生成Api文档
        /// </summary>
        /// <param name="config">
        /// {
        ///     "projectName": "项目名称",
        ///     "moduleName": "文档的模块名称",
        ///     "moduleTitle": "文档的标题",
        ///     "all": true/false "是否全部更新（按模块）"
        /// }
        /// </param>
        /// <returns></returns>
        [HttpPost]
        [Route("api/spec/ca")]
        public async Task<object> GenerateApiSpec()
        {
            try
            {
                using var reader = new StreamReader(Request.Body);
                var body = await reader.ReadToEndAsync();
                JObject config = JObject.Parse(body);

                // TODO: 可以考虑通过一个按钮，直接刷新当前文档。
                var helper = new SpecHelper(_repository.ConnectionManager.DbConnection.ReadConnectionString);
                var projectName = config.Value<string>("projectName");

                int count = 0;
                if (config.Value<bool?>("all") == true)
                {
                    var modules = helper.ListModules(projectName);

                    foreach (var item in modules)
                    {
                        var apis = helper.ListAPIForProject(projectName, item.moduleName);

                        var generator = new ApiSpecGenerator(projectName, item.moduleName, item.moduleTitle, apis);
                        count += generator.Generate(docRoot);
                    }
                }
                else
                {
                    var moduleName = config.Value<string>("moduleName");
                    var moduleTitle = config.Value<string>("moduleTitle");

                    var apis = helper.ListAPIForProject(projectName, moduleName);

                    var generator = new ApiSpecGenerator(projectName, moduleName, moduleTitle, apis);
                    count = generator.Generate(docRoot);
                }

                return new StandardResult { Data = new { count = count } };
            }
            catch(Exception ex)
            {
                return new StandardResult(500, "生成API文件失败: " + ex);
            }
        }

        [HttpGet]
        [Route("api/spec/list")]
        public object ListSpec(string projectName)
        {
            var result = new List<object>();

            var specDir = docRoot + projectName;
            if(Directory.Exists(specDir))
            {
                result.AddRange(Directory.GetFiles(specDir).Select(f => new { fileName = Path.GetFileNameWithoutExtension(f), type = "项目简介", filePath = f }));
            }

            var dbSpecDir = specDir + "/Database/";
            if (Directory.Exists(dbSpecDir))
            {
                result.AddRange(Directory.GetFiles(dbSpecDir).Select(f => new { fileName = Path.GetFileNameWithoutExtension(f), type = "数据库文档", filePath = f }));
            }

            var apiSpecDir = specDir + "/Api/";
            if (Directory.Exists(apiSpecDir))
            {
                result.AddRange(Directory.GetFiles(apiSpecDir).Select(f => new { fileName = Path.GetFileNameWithoutExtension(f), type = "接口文档", filePath = f }));
            }

            var otherSpecDir = specDir + "/Others/";
            if (Directory.Exists(otherSpecDir))
            {
                result.AddRange(Directory.GetFiles(otherSpecDir).Select(f => new { fileName = Path.GetFileNameWithoutExtension(f), type = "其它", filePath = f }));
            }

            return new StandardResult { Data = result };
        }

        [HttpPost("api/spec/import")]
        public object Import(string projectName)
        {
            if (string.IsNullOrEmpty(projectName)) return new StandardResult(ResultCode.LogicError, "No projectName specified.");
            if (Request.Form.Files.Count == 0) return new StandardResult(ResultCode.LogicError, "No file uploaded.");

            foreach (var file in Request.Form.Files)
            {
                using (StreamReader reader = new StreamReader(file.OpenReadStream()))
                {
                    var text = reader.ReadToEnd();
                    var dir = docRoot + projectName + "/Others/";
                    if (!Directory.Exists(dir)) Directory.CreateDirectory(dir);

                    var filePath = docRoot + projectName + "/Others/" + file.FileName;

                    System.IO.File.WriteAllText(filePath, text);
                }
            }

            return new StandardResult { Data = new { success = true } };
        }

        /// <summary>
        /// 删除指定路径的文档。
        /// </summary>
        /// <param name="filePath"></param>
        /// <returns></returns>
        [HttpDelete]
        [Route("api/spec/delete")]
        public object DeleteSpec(string filePath)
        {
            var result = new { success = System.IO.File.Exists(filePath) };

            if (System.IO.File.Exists(filePath))
            {
                System.IO.File.Delete(filePath);
            }

            return new StandardResult { Data = result };
        }

        public IActionResult Index([FromQuery]string project)
        {
            ViewData["projectName"] = project;

            var getStamp = new Func<string>(() => "?s="+DateTime.Now.Ticks);

            if (string.IsNullOrEmpty(project))
            {
                var projects = Directory.GetDirectories(docRoot).Select(p => Path.GetFileName(p));
                ViewData["projects"] = projects;
            }
            else
            {
                var dbSpecDir = docRoot + project + "/Database/";
                if (Directory.Exists(dbSpecDir))
                {
                    var dbSpecs = Directory.GetFiles(dbSpecDir);
                    ViewData["dbSpecs"] = dbSpecs.Select(f => f.Replace(docRoot, "doc/").Replace("\\", "/") + getStamp());
                }
                else
                {
                    ViewData["dbSpecs"] = new List<string>();
                }

                var apiSpecDir = docRoot + project + "/Api/";
                if (Directory.Exists(apiSpecDir))
                {
                    var apiSpecs = Directory.GetFiles(apiSpecDir);
                    ViewData["apiSpecs"] = apiSpecs.Select(f => f.Replace(docRoot, "doc/").Replace("\\", "/") + getStamp());
                }
                else
                {
                    ViewData["apiSpecs"] = new List<string>();
                }

                var otherSpecDir = docRoot + project + "/Others/";
                if (Directory.Exists(otherSpecDir))
                {
                    var otherSpecs = Directory.GetFiles(otherSpecDir);
                    ViewData["otherSpecs"] = otherSpecs.Select(f => f.Replace(docRoot, "doc/").Replace("\\", "/") + getStamp());
                }
                else
                {
                    ViewData["otherSpecs"] = new List<string>();
                }
            }

            return View();
        }        
    }
}
