﻿using Dapper;
using MySql.Data.MySqlClient;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.Data;
using System.Data.SqlClient;
using System.IO;
using System.Linq;
using Wicture.DbRESTFul.Configuration;
using Wicture.DbRESTFul.PMT.Models;
using Wicture.DbRESTFul.Resources.Csi;

namespace Wicture.DbRESTFul.PMT.Scaffolding
{
    public class CSIDbSource
    {
        public string projectName { get; set; }
        public string moduleName { get; set; }
        public List<DefaultACITable> tables { get; set; }

        public DatabaseType databaseType { get; set; }

        public IDbConnection GetConnection(string connectionString, DatabaseType DatabaseType = DatabaseType.MySQL)
        {
            if (DatabaseType == DatabaseType.SQLServer) return new SqlConnection(connectionString) as IDbConnection;
            else return new MySqlConnection(connectionString) as IDbConnection;
        }
    }


    public class ScaffoldingHelper
    {
        private readonly string databaseName = string.Empty;
        private readonly string query_table_names = "SELECT `TABLE_NAME` AS `name`, `TABLE_COMMENT` AS description FROM `INFORMATION_SCHEMA`.`TABLES` WHERE `TABLE_SCHEMA` = '{0}' AND `TABLE_TYPE` = 'BASE TABLE';";
        private readonly string query_table_names_sqlserver = "SELECT st.name,sep.VALUE  [description] FROM 	sys.tables st  LEFT JOIN sys.extended_properties sep ON st.object_id = sep.major_id AND sep.minor_id=0;";
        private readonly string query_all_column_names = "SELECT `TABLE_NAME` AS `table`, `COLUMN_NAME` AS `name`, `COLUMN_NAME` AS `title`, `COLUMN_TYPE` AS `type`, (CASE `IS_NULLABLE` WHEN 'NO' THEN 0 ELSE 1 END) AS nullable, (CASE `COLUMN_KEY` WHEN 'PRI' THEN 1 ELSE 0 END) AS primary_key, `COLUMN_DEFAULT` AS `default`,`COLUMN_COMMENT` AS description FROM `INFORMATION_SCHEMA`.`COLUMNS` WHERE `TABLE_SCHEMA` = '{0}';";
        private readonly string query_all_column_names_sqlserver = "SELECT st.name [table], sc.is_nullable nullable, sc.name [name], sc.name title, si.is_primary_key primary_key, sep.VALUE description, TYPE_NAME(sc.system_type_id) [type], 0 [default] FROM sys.tables st INNER JOIN sys.columns sc ON st.object_id = sc.object_id LEFT JOIN sys.extended_properties sep ON st.object_id = sep.major_id AND sc.column_id = sep.minor_id LEFT JOIN sys.sysindexkeys sik ON sik.id = sc.object_id AND sik.colid = sc.column_id LEFT JOIN sys.indexes si ON si.object_id = sik.id AND si.index_id = sik.indid LEFT JOIN sys.objects so ON so.name = si.name;";
        private readonly string query_api_for_project = "SELECT a.* FROM api AS a LEFT JOIN project AS p ON p.id = a.projectId WHERE p.name = @projectName;";
        private readonly string get_databaseType = "SELECT databaseType FROM project WHERE id=@projectId;";
        private static readonly List<string> ObjectProperties = new List<string> { "cache", "implementation", "parameter", "result", "mock" };

        private readonly CSIDbSource source;
        private readonly string pmtConncetString = "";
        private readonly string projectConncetString = "";

        public ScaffoldingHelper(string connectionString, int projectId = -1)
        {
            source = new CSIDbSource();
            projectConncetString = connectionString;
            using (var conn = source.GetConnection(connectionString))
            {
                databaseName = conn.Database;
            }

            pmtConncetString = ConfigurationManager.Settings.GetConfig<CsiSectionConfig>("CSI").DbConnection.ReadConnectionString;
            if (projectId != -1)
            {
                try
                {
                    using (var conn = source.GetConnection(pmtConncetString))
                    {
                        var databaseType = SqlMapper.Query<int?>(conn, get_databaseType, new { projectId = projectId }).FirstOrDefault();
                        source.databaseType = databaseType == 1 ? DatabaseType.SQLServer : DatabaseType.MySQL;
                    }
                }
                catch (Exception ex)
                {
                    LoggerManager.Logger.Error(ex, "Get database type failed.");
                    throw ex;
                }
            }
        }

        public List<DefaultACITable> ListTables()
        {
            try
            {
                using (var conn = source.GetConnection(projectConncetString, source.databaseType))
                {
                    if (conn.State == ConnectionState.Closed) conn.Open();
                    //choose sqls
                    var query_all_columns = source.databaseType == DatabaseType.MySQL ? query_all_column_names : query_all_column_names_sqlserver;
                    var query_all_tables = source.databaseType == DatabaseType.MySQL ? query_table_names : query_table_names_sqlserver;
                    //get columns
                    var columns = SqlMapper.Query<DefaultACIColumn>(conn, string.Format(query_all_columns, databaseName)).ToList();
                    //get tables
                    var data = SqlMapper.Query<DefaultACITable>(conn, string.Format(query_all_tables, databaseName));
                    //make result for return
                    data.ForEach(t =>
                    {
                        t.title = t.description;
                        t.columns = columns.Where(c => c.table == t.name).ToList();
                    });
                    return data.ToList();
                }
            }
            catch (Exception ex)
            {
                LoggerManager.Logger.Error(ex, "Get list tables failed.");
                throw new Exception("获取项目数据库表信息出错。", ex);
            }
        }

        public List<JObject> ListAPIForProject(string projectName, string moduleName)
        {
            var result = new List<JObject>();
            try
            {
                using (var conn = source.GetConnection(pmtConncetString))
                {
                    var param = new { projectName = projectName };
                    IEnumerable<dynamic> apis = SqlMapper.Query(conn, query_api_for_project, param);

                    foreach (var item in apis)
                    {
                        // If moduleName is not specified, list all.
                        if (string.IsNullOrEmpty(moduleName) || item.module == moduleName)
                        {
                            JObject ca = JObject.FromObject(item);
                            ObjectProperties.ForEach(p => ca[p] = JToken.Parse(ca.Value<string>(p) ?? "{}"));
                            result.Add(ca);
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                LoggerManager.Logger.Error(ex, "Get list api for project failed.");
                throw new Exception("获取项目接口信息出错。", ex);
            }

            return result;
        }

        public void GenerateCSI(DefaultACITable dACITable)
        {
            var TableName = char.ToUpper(dACITable.name[0]) + dACITable.name.Substring(1);
            var tableName = dACITable.name;
            var primaryKey = dACITable.columns.FirstOrDefault(c => c.primary_key)?.name;
            var columnNames = string.Join(", ", dACITable.columns.Select(c => c.name));
            var columnValues = string.Join(", ", dACITable.columns.Select(c => "@" + c.name));
            var nameValues = string.Join(", ", dACITable.columns.Select(c => $"{c.name} = @{c.name}"));

            var text = File.ReadAllText(Path.Combine(Directory.GetCurrentDirectory(), "Template//template.json"));

            var result = text.Replace("#TableName#", TableName)
                            .Replace("#tableName#", tableName)
                            .Replace("#primaryKey#", primaryKey)
                            .Replace("#columnNames#", columnNames)
                            .Replace("#columnValues#", columnValues)
                            .Replace("#nameValues#", nameValues);
        }
    }
}
