﻿using Microsoft.IdentityModel.Tokens;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.IdentityModel.Tokens.Jwt;
using System.Linq;
using System.Security.Claims;

namespace Wicture.DbRESTFul.Authentication
{
    public class DefaultJwtTokenConverter : ITokenConverter
    {
        protected const string DateTimeStringFormat = "yyyy-MM-dd HH:mm:ss";
        protected readonly JwtSecurityTokenHandler handler = new JwtSecurityTokenHandler();

        private readonly ITokenExpirationStratagem _tokenExpirationStratagem;
        private readonly IIdentityStateManager _tokenValidator;
        private readonly TokenOptions _tokenOptions;
            
        public DefaultJwtTokenConverter(
            ITokenExpirationStratagem tokenExpirationStratagem, 
            ITokenOptionsProvider tokenOptionsProvider, 
            IIdentityStateManager tokenValidator)
        {
            _tokenExpirationStratagem = tokenExpirationStratagem;
            _tokenValidator = tokenValidator;
            _tokenOptions = tokenOptionsProvider.Provide();
        }

        public virtual UserIdentity Parse(string token)
        {
            SecurityToken secToken;
            var pricipal = handler.ValidateToken(token, _tokenOptions.ValidationParameters, out secToken);

            var userIdclaim = pricipal.Claims.FirstOrDefault(c => c.Type == "Id");
            var userNameclaim = pricipal.Claims.FirstOrDefault(c => c.Type == "Name");
            var roleclaim = pricipal.Claims.FirstOrDefault(c => c.Type == "Role");
            var aliasclaim = pricipal.Claims.FirstOrDefault(c => c.Type == "Alias");
            var userDataclaim = pricipal.Claims.FirstOrDefault(c => c.Type == "Properties");
            var iatClaim = pricipal.Claims.FirstOrDefault(c => c.Type == JwtRegisteredClaimNames.Iat);

            var properties = string.IsNullOrEmpty(userDataclaim?.Value) ? null : JsonConvert.DeserializeObject<Dictionary<string, object>>(userDataclaim.Value);
            var identity = new UserIdentity()
            {
                AuthenticationType = "JwtBearer",
                Id = userIdclaim?.Value ?? "",
                Name = userNameclaim?.Value ?? "",
                Role = roleclaim?.Value ?? "",
                Alias = aliasclaim?.Value ?? "",
                Properties = properties,
                ExpiresUtc = secToken?.ValidTo.ToString(DateTimeStringFormat),
                IssuedUtc = iatClaim?.Value
            };

            return identity;
        }

        public virtual JObject Compose(UserIdentity identity)
        {
            var now = DateTime.UtcNow;
            var nowUnixLongTime = now.GetUnixLongTime();

            var claims = new List<Claim>
            {
                new Claim("Id", identity?.Id ?? ""),
                new Claim("Name", identity?.Name ?? ""),
                new Claim(JwtRegisteredClaimNames.Jti, Guid.NewGuid().ToString()),
                new Claim(JwtRegisteredClaimNames.Iat, nowUnixLongTime.ToString(), ClaimValueTypes.Integer64)
            };

            if (!string.IsNullOrEmpty(identity?.Role)) claims.Add(new Claim("Role", identity.Role));
            if (!string.IsNullOrEmpty(identity?.Alias)) claims.Add(new Claim("Alias", identity.Alias));
            if (identity?.Properties?.Count > 0) claims.Add(new Claim("Properties", JsonConvert.SerializeObject(identity.Properties)));

            var expires = _tokenExpirationStratagem.GetExpirationSpan(_tokenOptions, identity);

            var jwt = new JwtSecurityToken(
                issuer: _tokenOptions.Issuer,
                audience: _tokenOptions.Audience,
                claims: claims,
                notBefore: now,
                expires: now.Add(expires),
                signingCredentials: _tokenOptions.SigningCredentials);
            var encodedJwt = new JwtSecurityTokenHandler().WriteToken(jwt);

            JObject token = new JObject();
            token[".issued"] = nowUnixLongTime;
            token[".expires"] = now.Add(expires).GetUnixLongTime();
            token["id"] = identity.Id;
            token["name"] = identity.Name;
            token["role"] = identity.Role;
            token["alias"] = identity.Alias;
            token["expires_in"] = (expires).TotalMilliseconds;
            token["access_token"] = encodedJwt;
            token["token_type"] = "bearer";

            if (identity.Properties != null)
            {
                foreach (var item in identity.Properties)
                {
                    token[item.Key] = item.Value?.ToString();
                }
            }

            _tokenValidator.UpdateStateAsync(identity, expires);

            return token;
        }

    }
}
