﻿using Microsoft.Extensions.Logging;
using System;
using System.Threading.Tasks;

namespace Wicture.DbRESTFul.Authentication
{
    public class ExclusiveIdentityStateManager : IIdentityStateManager
    {
        private readonly IIdentityStateStorageService _tokenStorageService;
        private readonly ILogger _logger;

        public bool Enabled => false;

        public ExclusiveIdentityStateManager(ILoggerFactory loggerFactory, 
            IIdentityStateStorageService tokenStorageService)
        {
            _logger = loggerFactory.CreateLogger("DbRESTFulApi");
            _tokenStorageService = tokenStorageService;
        }
        
        public async Task ValidateStateAsync(UserIdentity identity)
        {
            if (!Enabled) return;

            var key = GetCacheTokenKey(identity);
            var tokenStamp = await _tokenStorageService.GetAsync(key);

            if (!string.IsNullOrEmpty(tokenStamp) && tokenStamp == identity.IssuedUtc) return;

            if (string.IsNullOrEmpty(tokenStamp))
            {
                _logger.LogInformation($"Unable to get value for key `{tokenStamp}` for member: {identity.Name}");
                throw new LogicalException($"Token was invalid (expired or by InvalidateToken) for user: {identity.Id} from: {identity.From}", ResultCode.TokenRefreshedStatusCode);
            }

            if (tokenStamp != identity.IssuedUtc)
            {
                _logger.LogInformation($"token was renewed for member `{identity.Name}`");
                throw new LogicalException($"Token was invalid (refresh or login in other device.) for user: {identity.Id} from: {identity.From}", ResultCode.TokenRefreshedStatusCode);
            }
        }

        public async Task InvalidateStateAsync(UserIdentity identity, string source)
        {
            if (!Enabled) throw new NotSupportedException("TokenValidator is not enabled.");

            var key = GetCacheTokenKey(identity);
            await _tokenStorageService.RemoveAsync(key);
            _logger.LogInformation($"Token for member: {identity.Name} was removed by {source}.");
        }

        public async Task UpdateStateAsync(UserIdentity identity, TimeSpan expiration)
        {
            if (Enabled)
            {
                var key = GetCacheTokenKey(identity);
                await _tokenStorageService.SetAsync(key, identity.IssuedUtc, (int)expiration.TotalSeconds);
            }
        }

        protected virtual string GetCacheTokenKey(UserIdentity identity)
        {
            object loginFrom = null;
            bool? hasLoginFrom = identity.Properties?.TryGetValue("loginFrom", out loginFrom);
            if (hasLoginFrom == true && loginFrom?.ToString().ToUpper() == "APP")
            {
                return $"Token_{identity.Id}_{identity.Name}_APP";
            }
            if (hasLoginFrom == true && loginFrom?.ToString().ToUpper() == "WECHAT")
            {
                return $"Token_{identity.Id}_{identity.Name}_WECHAT";
            }
            return $"Token_{identity.Id}_{identity.Name}";
        }
    }
}
