﻿using Microsoft.AspNetCore.Http;
using Microsoft.Extensions.Logging;
using System;
using System.Threading.Tasks;
using Wicture.DbRESTFul.Resources.Api;

namespace Wicture.DbRESTFul.Authentication
{
    public class JwtTokenAuthenticationService : IApiAuthenticationService
    {
        private ILogger _logger;
        private ITokenConverter _tokenParser;
        private IIdentityStateManager _tokenValidator;

        public JwtTokenAuthenticationService(ILoggerFactory loggerFactory, 
            ITokenConverter tokenParser, 
            IIdentityStateManager tokenValidator)
        {
            _logger = loggerFactory.CreateLogger("DbRESTFulApi");
            _tokenParser = tokenParser;
            _tokenValidator = tokenValidator;
        }

        public virtual async Task<UserIdentity> Authenticate(HttpContext context, ApiModel api)
        {
            var token = context.Request.Headers["Authorization"].ToString();
            var from = context.Request.Headers["From"].ToString();

            return await Authenticate(token, api, from);
        }

        public virtual async Task<UserIdentity> Authenticate(string token, ApiModel api, string from = "rpc")
        {
            if (string.IsNullOrEmpty(token) || !token.ToLower().StartsWith("bearer") || token.Length <= TokenOptions.StartIndex)
            {
                token = "";
            }
            else
            {
                token = token.Substring(TokenOptions.StartIndex);
            }
                
            if (string.IsNullOrEmpty(token) || token.Equals("null"))
            {
                if (api.allowAnonymous)
                {
                    var identity = UserIdentity.Anonymous;
                    identity.From = from;
                    return identity;
                }

                throw new LogicalException($"请重新登录", ResultCode.TokenRequiredStatusCode);
            }

            try
            {
                var identity = _tokenParser.Parse(token);
                await _tokenValidator.ValidateStateAsync(identity);
                return identity;
            }
            catch (Exception ex)
            {
                if (api.allowAnonymous)
                {
                    var identity = UserIdentity.Anonymous;
                    identity.From = from;
                    return identity;
                }
                _logger.LogError(ex, $"Unauthorized, Token validation failed. Token: `{token}`");
                throw ex is LogicalException ? ex : new LogicalException($"请重新登录", ResultCode.UnauthorizedStatusCode);
            }
        }
    }
}
