﻿using Microsoft.Extensions.Caching.Memory;
using Microsoft.Extensions.Logging;
using Newtonsoft.Json.Linq;
using System;
using System.Threading.Tasks;

namespace Wicture.DbRESTFul.Cache
{
    /// <summary>
    /// Memory 缓存实现。
    /// </summary>
    public class MemoryCacheProvider : ICacheProvider
    {
        private readonly IMemoryCache _memoryCache;
        private readonly ILogger _logger;

        public event EventHandler OnError;

        public string Name { get; } = "memory";

        public MemoryCacheProvider(IMemoryCache memoryCache, ILoggerFactory loggerFactory)
        {
            _memoryCache = memoryCache;
            _logger = loggerFactory.CreateLogger("DbRESTFul");
        }

        public async Task Set(string key, object data, int expirationInSeconds)
        {
            try
            {
                _memoryCache.Set(key, data, DateTime.UtcNow.AddSeconds(expirationInSeconds));
                await Task.CompletedTask;
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Set Cache failed for key: `{key}`");
                OnError?.Invoke(this, null);
            }
        }

        public async Task Set(string key, object data, DateTime expiresAt)
        {
            try
            {
                _memoryCache.Set(key, data, expiresAt);
                await Task.CompletedTask;
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Set Cache failed for key: `{key}`");
                OnError?.Invoke(this, null);
            }
        }

        public async Task SetString(string key, string data, int expirationInSeconds)
        {
            await Set(key, data, expirationInSeconds);
        }

        public async Task SetString(string key, string data, DateTime expiresAt)
        {
            await Set(key, data, expiresAt);
        }

        public async Task<ulong> Increase(string key, DateTime expiresAt, ulong defaultValue = 1, ulong delta = 1)
        {
            try
            {
                var counter = _memoryCache.GetOrCreate<ulong>(key, a => defaultValue);
                if(counter > defaultValue)
                {
                    _memoryCache.Set(key, counter + delta, expiresAt);
                }

                await Task.CompletedTask;
                return counter;
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Set Cache failed for key: `{key}`");
                OnError?.Invoke(this, null);
                return 0;
            }
        }

        public async Task<ulong> GetIncr(string key, bool del = true)
        {
            if (_memoryCache.TryGetValue(key, out ulong counter))
            {
                _memoryCache.Remove(key);
            }

            await Task.CompletedTask;
            return counter;
        }

        public async Task<JToken> Get(string key)
        {
            try
            {
                var result = _memoryCache.Get(key);
                if (result == null) return null;
                return await Task.FromResult((result is JToken) ? (result as JToken) : JToken.FromObject(result));
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Get Cache failed for key: `{key}`");
                OnError?.Invoke(this, null);
                return null;
            }
        }

        public async Task<string> GetString(string key)
        {
            try
            {
                var result = _memoryCache.Get(key);
                if (result == null) return null;
                return await Task.FromResult((result is string) ? (result as string) : result.ToString());
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Get Cache failed for key: `{key}`");
                OnError?.Invoke(this, null);
                return null;
            }
        }

        public async Task Set(string key, object data)
        {
            try
            {
                _memoryCache.Set(key, data);
                await Task.CompletedTask;
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Set Cache failed for key: `{key}`");
                OnError?.Invoke(this, null);
            }
        }

        public async Task SetString(string key, string data)
        {
            await Set(key, data);
        }

        public async Task Remove(params string[] key)
        {
            try
            {
                foreach (var _key in key)
                {
                    _memoryCache.Remove(_key);
                }
                await Task.CompletedTask;
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Remove Cache failed for key: `{key}`");
                OnError?.Invoke(this, null);
            }
        }

        public IMemoryCache MemoryCache { get { return _memoryCache; } }
    }
}
