﻿using CSRedis;
using Microsoft.Extensions.Caching.Distributed;
using Microsoft.Extensions.Caching.Redis;
using Microsoft.Extensions.Logging;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using System;
using System.Linq;
using System.Threading.Tasks;
using Wicture.DbRESTFul.Configuration;

namespace Wicture.DbRESTFul.Cache
{
    /// <summary>
    /// Redis 缓存实现。
    /// </summary>
    public class RedisCacheProvider : ICacheProvider
    {
        private CSRedisClient client;
        private readonly ISectionConfigProvider _sectionConfigProvider;
        private readonly ILogger _logger;
        private const int RetryCount = 3;

        public event EventHandler OnError;

        public RedisCacheProvider(ILoggerFactory loggerFactory, ISectionConfigProvider sectionConfigProvider)
        {
            _sectionConfigProvider = sectionConfigProvider;
            _logger = loggerFactory.CreateLogger("DbRESTFul");

            client = Create();
            OnError += (s, e) => Reinit();
        }

        public string Name { get; } = "redis";

        public CSRedisClientLock Lock(string name, int timeoutSeconds, bool autoDelay = true) {
            try {
                return client.Lock(name, timeoutSeconds, autoDelay);
            } catch (Exception ex) {
                _logger.LogError(ex, $"distribute lock failed for key: `{name}`");
                OnError?.Invoke(this, null);
                return null;
            }
        }

        public async Task Set(string key, object data)
        {
            try
            {
                var val = JsonConvert.SerializeObject(data);
                await client.SetAsync(key, val);
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Set Cache failed for key: `{key}`");
                OnError?.Invoke(this, null);
            }
        }

        public async Task Set(string key, object data, int expiration)
        {
            await SetString(key, JsonConvert.SerializeObject(data), expiration);
        }

        public async Task Set(string key, object data, DateTime expiresAt)
        {
            await SetString(key, JsonConvert.SerializeObject(data), expiresAt);
        }

        public async Task SetString(string key, string data)
        {
            try
            {
                await client.SetAsync(key, data);
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Set Cache failed for key: `{key}`");
                OnError?.Invoke(this, null);
            }
        }

        public async Task SetString(string key, string data, int expiration)
        {
            try
            {
                await client.SetAsync(key, data, expiration);
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Set Cache failed for key: `{key}`");
                OnError?.Invoke(this, null);
            }
        }

        public async Task SetString(string key, string data, DateTime expiresAt)
        {
            try
            {
                await client.SetAsync(key, data);
                await client.ExpireAtAsync(key, expiresAt);
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Set Cache failed for key: `{key}`");
                OnError?.Invoke(this, null);
            }
        }

        public async Task<bool> KeyExist(string key)
        {
            try
            {
                return await client.ExistsAsync(key);
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Check KeyExist failed for key: `{key}`");
                OnError?.Invoke(this, null);
                return false;
            }
        }

        public async Task SetExpiration(string key, int expiration)
        {
            try
            {
                await client.ExpireAsync(key, expiration);
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Set Expiration failed for key: `{key}`");
                OnError?.Invoke(this, null);
            }
        }

        public async Task SetExpiration(string key, DateTime expireAt)
        {
            try
            {
                await client.ExpireAtAsync(key, expireAt);
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Set Expiration failed for key: `{key}`");
                OnError?.Invoke(this, null);
            }
        }

        public async Task<ulong> Increase(string key, DateTime expiresAt = default(DateTime), ulong defaultValue = 1, ulong delta = 1)
        {
            try
            {
                // 如果key不存在，则第一次创建的时候，设置其过期时间
                if (expiresAt != default(DateTime) && !(await client.ExistsAsync(key)))
                {
                    var result = (ulong)(await client.IncrByAsync(key, (long)defaultValue));
                    await client.ExpireAtAsync(key, expiresAt);
                    return result;
                }

                return (ulong)(await client.IncrByAsync(key, (long)defaultValue));
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Increase failed for key: `{key}`");
                OnError?.Invoke(this, null);
                return 0;
            }
        }

        public async Task<JToken> Get(string key)
        {
            try
            {
                var data = await client.GetAsync(key);
                return string.IsNullOrEmpty(data) ? null : JToken.Parse(data);
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Get Cache failed for key: `{key}`");
                OnError?.Invoke(this, null);

                // 重试
                var data = await RetryGet(key);
                return string.IsNullOrEmpty(data) ? null : JToken.Parse(data);
            }
        }

        public async Task<string> GetString(string key)
        {
            try
            {
                return client.Get(key);
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Get Cache failed for key: `{key}`");
                OnError?.Invoke(this, null);

                // 重试
                return await RetryGet(key);
            }
        }

        public async Task<T> Get<T>(string key)
        {
            try
            {
                var data = await client.GetAsync(key);
                return string.IsNullOrEmpty(data) ? default(T) : JsonConvert.DeserializeObject<T>(data);
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Get Cache failed for key: `{key}`");
                OnError?.Invoke(this, null);

                // 重试
                return await RetryGet<T>(key);
            }
        }

        private async Task<T> RetryGet<T>(string key, int retry = 1)
        {
            try
            {
                var data = await client.GetAsync(key);
                _logger.LogInformation($"Retry to get Cache successfully for key: `{key}` with temp: {retry}.");
                return string.IsNullOrEmpty(data) ? default(T) : JsonConvert.DeserializeObject<T>(data);
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Retry to get Cache failed for key: `{key}` with temp: {retry}");
                OnError?.Invoke(this, null);

                if (retry >= RetryCount) return default(T);
                return await RetryGet<T>(key, retry + 1);
            }
        }

        private async Task<string> RetryGet(string key, int retry = 1)
        {
            try
            {
                var data = await client.GetAsync(key);
                _logger.LogInformation($"Retry to get Cache successfully for key: `{key}` with temp: {retry}.");
                return data;
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Retry to get Cache failed for key: `{key}` with temp: {retry}");
                OnError?.Invoke(this, null);

                if (retry >= RetryCount) return null;
                return await RetryGet(key, retry + 1);
            }
        }

        public async Task Remove(params string[] key)
        {
            try
            {
                await client.DelAsync(key);
            }
            catch (Exception ex)
            {
                _logger.LogError(ex, $"Remove Cache failed for key: `{key}`");
                OnError?.Invoke(this, null);
            }
        }

        private void Reinit()
        {
            try
            {
                var cache = Create();
                if (cache == null)
                {
                    _logger.LogError($"Create cache failed due to the ConnectionString issue.");
                    return;
                }

                client = cache;
                _logger.LogInformation($"Force reinit for Redis Cache.");
            }
            catch (Exception ex)
            {
                _logger.LogError($"Force reinit failed. {ex}");
            }
        }

        public async Task<ulong> GetIncr(string key, bool del = true)
        {
            await Task.CompletedTask;
            throw new NotImplementedException();
        }

        public CSRedisClient CSRedisClient { get { return client; } }

        private CSRedisClient Create()
        {
            var config = _sectionConfigProvider.Provide<CacheSectionConfig>()?.GetRedisConfig();

            if (string.IsNullOrEmpty(config?.ConnectionString))
            {
                _logger.LogWarning($"No ConnectionString for Redis.");
                return null;
            }

            CSRedisClient csredis = null;
            if (config?.Sentinels != null)
            {
                csredis = new CSRedisClient(config.ConnectionString, config.Sentinels);
            }
            else
            {
                var connectionStrings = config.ConnectionString.Split(";".ToArray(), StringSplitOptions.RemoveEmptyEntries);
                csredis = new CSRedisClient(null, connectionStrings);
            }

            RedisHelper.Initialization(csredis);
            return csredis;
        }
    }
}
