﻿using Newtonsoft.Json;
using Newtonsoft.Json.Converters;
using Newtonsoft.Json.Linq;
using Newtonsoft.Json.Serialization;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net.Http;
using System.Reflection;
using System.Threading.Tasks;
using Wicture.DbRESTFul.Configuration;
using Wicture.DbRESTFul.Resources;
using Wicture.DbRESTFul.Resources.Api;
using Wicture.DbRESTFul.Resources.Csi;

namespace Wicture.DbRESTFul.DevTool
{
    public class ResourceSynchronizer
    {
        private readonly ApiSectionConfig _apiSectionConfig;
        private readonly CsiSectionConfig _csiSectionConfig;
        private readonly DevToolSectionConfig _devToolConfig;
        private readonly IResourceManager _resourceManager;

        public ResourceSynchronizer(ISectionConfigProvider sectionConfigProvider, IResourceManager resourceManager)
        {
            _devToolConfig = sectionConfigProvider.Provide<DevToolSectionConfig>();

            if (!string.IsNullOrEmpty(_devToolConfig.PMTServiceUrl))
            {
                _devToolConfig.PMTServiceUrl = _devToolConfig.PMTServiceUrl.TrimEnd('/');
            }

            _apiSectionConfig = sectionConfigProvider.Provide<ApiSectionConfig>();
            _csiSectionConfig = sectionConfigProvider.Provide<CsiSectionConfig>();
            _resourceManager = resourceManager;
        }

        public async Task<List<SyncApiModule>> ListModulesAsync()
        {
            var url = $"{_devToolConfig.PMTServiceUrl}/dbrestful/devtool/api/modules";
            var result = await GetResource<List<SyncApiModule>>(url, new { _devToolConfig.ProjectName });

            if (result.StatusCode == 200)
            {
                return result.Data;
            }
            else
            {
                throw new LogicalException(result.ErrorMessage, 500);
            }
        }

        public async Task<object> SyncApisAsync(string projectId, string moduleId)
        {
            var url = $"{_devToolConfig.PMTServiceUrl}/dbrestful/devtool/api/sync";
            var result = await GetResource<JObject>(url, new { projectId, moduleId, branch = _devToolConfig.Branch });

            if (result.StatusCode != 200) throw new LogicalException(result.ErrorMessage, 500);

            if (!Directory.Exists(_apiSectionConfig.Path))
            {
                Directory.CreateDirectory(_apiSectionConfig.Path);
            }

            foreach (var item in result.Data)
            {
                var path = string.Format("{0}/{1}.json", _apiSectionConfig.Path, item.Key);
                LoggerManager.Logger.Info($"Api was synced for {item.Key}.");

                var apis = item.Value.ToObject<List<ApiModel>>();
                JsonSerializerSettings settings = new JsonSerializerSettings
                {
                    Converters = new JsonConverter[] { new StringEnumConverter(new CamelCaseNamingStrategy()) },
                    Formatting = Formatting.Indented
                };
                File.WriteAllText(path, JsonConvert.SerializeObject(apis, settings));
            }

            _resourceManager.LoadApi();

            return new { succeeded = true };
        }

        public async Task<object> SyncCsisAsync(string projectId, string moduleId)
        {
            var csiPath = _csiSectionConfig?.Path;
            if (string.IsNullOrEmpty(csiPath))
            {
                throw new InvalidOperationException("Csi is not configured or supported.");
            }

            csiPath = Path.Combine(Directory.GetCurrentDirectory(), csiPath);
            if (!Directory.Exists(csiPath))
            {
                Directory.CreateDirectory(csiPath);
            }

            var url = $"{_devToolConfig.PMTServiceUrl}/dbrestful/devtool/csi/sync";
            var result = await GetResource<JObject>(url, new { projectId, moduleId, branch = _devToolConfig.Branch });

            if (result.StatusCode != 200) throw new LogicalException(result.ErrorMessage, 500);

            foreach (var item in result.Data)
            {
                item.Value.ForEach(a =>
                {
                    var dict = a as IDictionary<string, JToken?>;
                    if (dict.ContainsKey("middleWares")) // Compatible with earlier version of PMT.
                    {
                        a["middlewares"] = a["middleWares"];
                        a.RemoveProperties("middleWares");
                    }

                    if (a["middlewares"].Type == JTokenType.String)
                    {
                        var middlewares = a.Value<string>("middlewares");
                        if (middlewares == "null") middlewares = "{}";
                        a["middlewares"] = string.IsNullOrEmpty(middlewares) ? null : JObject.Parse(middlewares);
                    }
                });

                var csis = item.Value.ToObject<List<CsiModel>>();

                foreach (var csi in csis)
                {
                    if (csi.code.Type == JTokenType.String && csi.code.ToString().Contains("{"))
                    {
                        csi.code = JObject.Parse(csi.code.ToString());
                    }
                }

                File.WriteAllText($"{csiPath}/{item.Key}.json", JsonConvert.SerializeObject(csis, Formatting.Indented));
                LoggerManager.Logger.Info($"Csi was synced for {item.Key}.");
            }

            _resourceManager.LoadCsi();

            return new { succeeded = true };
        }

        private async Task<StandardResult<T>> GetResource<T>(string url, object param)
        {
            try
            {
                using (HttpClient client = new HttpClient())
                {
                    var requestUri = url += "?" + Parameterize(param);
                    var response = await client.GetAsync(requestUri);
                    response.EnsureSuccessStatusCode();

                    var dataAsString = await response.Content.ReadAsStringAsync();
                    var result = JToken.Parse(dataAsString);

                    return result.Type == JTokenType.Object && result["statusCode"] != null
                        ? result.ToObject<StandardResult<T>>()
                        : new StandardResult<T> { StatusCode = 200, Data = result.ToObject<T>() };
                }
            }
            catch (Exception ex)
            {
                throw ex;
            }
        }

        private static string Parameterize(object param)
        {
            var properties = param.GetType().GetTypeInfo().GetProperties(BindingFlags.Public | BindingFlags.FlattenHierarchy | BindingFlags.Instance);
            var result = properties.Select(pi => $"{pi.Name}={pi.GetValue(param)}");

            return string.Join("&", result);
        }
    }
}
