﻿using Autofac;
using Newtonsoft.Json.Linq;
using System.Linq;
using System.Reflection;
using System.Threading.Tasks;
using Wicture.DbRESTFul.Infrastructure.Repository;
using Wicture.DbRESTFul.Resources;
using Wicture.DbRESTFul.Resources.Api;
using Wicture.DbRESTFul.Resources.Repository;

namespace Wicture.DbRESTFul.Infrastructure.Executors.Api
{
    public class RepositoryExecutor : IExecutor
    {
        public ImplementationType ImplementationType => ImplementationType.Repository;

        private readonly RepositoryLoader _repositoryLoader;
        public RepositoryExecutor(RepositoryLoader repositoryLoader)
        {
            _repositoryLoader = repositoryLoader;
        }

        public async Task<object> ExecuteAsync(ExecutionContext context)
        {
            var key = context.Api.implementation.name;
            if (!_repositoryLoader.Items.ContainsKey(key))
            {
                throw new DbRESTFulException($"The repository implementation '{key}' is not found.");
            }

            using (var scope = AutofacContainer.Container.BeginLifetimeScope())
            {
                var item = _repositoryLoader.Items[key];

                var repository = scope.Resolve(item.Type) as IDbRESTFulRepository;
                repository.ConnectionManager.Resolve(context);
                repository.Context = context;
                repository.ResourceManager = scope.Resolve<IResourceManager>();

                return await InvokeAsync(repository, item, context);
            }
        }

        private async Task<object> InvokeAsync(IDbRESTFulRepository repository, RepositoryModel repositoryModel, ExecutionContext context)
        {
            try
            {
                var parameterInfo = repositoryModel.Method.GetParameters().FirstOrDefault();
                object parameter = parameterInfo != null ? context.Param : null;
                if (parameterInfo != null && !parameterInfo.ParameterType.IsAssignableFrom(typeof(JToken)))
                {
                    parameter = context.Param?.ToObject(parameterInfo.ParameterType);
                }

                var result = parameterInfo == null
                    ? repositoryModel.Method.Invoke(repository, null)
                    : repositoryModel.Method.Invoke(repository, new object[] { parameter });

                if (result is Task)
                {
                    var task = result as Task;
                    await task.ConfigureAwait(false);
                    return (object)((dynamic)task).Result;
                }

                return result;
            }
            catch (TargetInvocationException ex)
            {
                // 因为通过反射调用时，所有方法抛出的异常会被包在TargetInvocationException的InnerException里。
                if (ex.InnerException != null && ex.InnerException is LogicalException)
                {
                    throw ex.InnerException;
                }
                else
                {
                    throw ex;
                }
            }
        }
    }
}
