﻿using Microsoft.Extensions.Logging;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using System;
using System.Linq;
using System.Threading.Tasks;
using Wicture.DbRESTFul.Configuration;
using Wicture.DbRESTFul.Resources.Api;

namespace Wicture.DbRESTFul.Infrastructure.Filters
{
    public class DefaultApiFilter : IApiFilter
    {
        private readonly ILogger _logger;

        private readonly ApiLogConfig _logConfig; 
        
        public DefaultApiFilter(ILoggerFactory loggerFactory, ISectionConfigProvider sectionConfigProvider)
        {
            _logConfig = sectionConfigProvider.Provide<ApiSectionConfig>().LogConfig;
            _logger = loggerFactory.CreateLogger("DbRESTFulApi");
        }

        public async Task OnExecutingAsync(ExecutionContext context)
        {
            _logger.LogInformation($"Executing api: `{context.Api.name}` started at: {context.StartedAt}");
            _logger.LogInformation(BuildLogEntry(context, true).ToString(_logConfig.WellFormat ? Formatting.Indented : Formatting.None));
            await Task.CompletedTask;
        }

        public async Task OnExecutedAsync(ExecutionContext context)
        {
            _logger.LogInformation($"Executed api: `{context.Api.name}`.");
            _logger.LogInformation(BuildLogEntry(context, false).ToString(_logConfig.WellFormat ? Formatting.Indented : Formatting.None));
            await Task.CompletedTask;
        }

        private JObject BuildLogEntry(ExecutionContext context, bool executing)
        {
            JObject logEntry;
            if (executing && (context is HttpApiExecutionContext httpExecutionContext))
            {
                logEntry = JObject.FromObject(new
                {
                    Request = new
                    {
                        Path = httpExecutionContext.HttpContext.Request.Path.Value,
                        Ip = httpExecutionContext.HttpContext.GetRemoteIpAddressBehindProxy(),
                        Agent = httpExecutionContext.HttpContext.Request.Headers["User-Agent"].FirstOrDefault()
                    }
                });

                var token = httpExecutionContext.HttpContext.Request.Headers.ContainsKey("Authorization") 
                            ? httpExecutionContext.HttpContext.Request.Headers["Authorization"].ToString()
                            : string.Empty;
                if (_logConfig.Token && token.Length > 0)
                {
                    logEntry["Token"] = token;
                }
            }
            else
            {
                logEntry = new JObject();
            }

            if (executing)
            {
                if (_logConfig.Param && context?.Param != null) logEntry["Param"] = context?.Param;
                if (_logConfig.Identity && context?.Identity != null) logEntry["Identity"] = JValue.FromObject(DesensitizedIdentity(context?.Identity));
            }
            else
            {
                if (_logConfig.Result && context?.ExecutionResult != null) logEntry["Result"] = JValue.FromObject(context?.ExecutionResult);
                if (_logConfig.Exception && context.Exception != null)
                {
                    logEntry["Exception"] = context?.Exception.ToString();
                }
            }

            return logEntry;
        }

        // Desensitizes Identity.
        private object DesensitizedIdentity(UserIdentity identity)
        {
            return identity == null ? (object)"" : new
            {
                identity.Id,
                identity.Name,
                identity.Role,
                identity.Alias,
                identity.From
            };
        }
    }
}
