﻿using Newtonsoft.Json;
using System.Threading.Tasks;
using Wicture.DbRESTFul.Configuration;
using Wicture.DbRESTFul.Resources.Api;

namespace Wicture.DbRESTFul.Infrastructure.Filters
{
    public class DefaultResultFilter : IResultFilter
    {
        private readonly ApiSectionConfig _apiConfig;
        private readonly IResultFormater _resultFormater;

        public DefaultResultFilter(IResultFormater resultFormater, ISectionConfigProvider sectionConfigProvider)
        {
            _apiConfig = sectionConfigProvider.Provide<ApiSectionConfig>();
            _resultFormater = resultFormater;
        }

        public virtual async Task OnHttpApiResultExecuting(HttpApiExecutionContext context)
        {
            context.HttpContext.Response.Headers.Add("Content-Type", "application/json");
            await Task.CompletedTask;
        }

        public virtual async Task OnHttpApiResultComposition(HttpApiExecutionContext context)
        {
            var statusCode = context switch
            {
                var c when c.Exception == null => ResultCode.OK,
                var c when c.Exception is LogicalException ex => ex.ErrorCode,
                var c when c.Exception is DbRESTFulException ex => ex.ErrorCode,
                _ => ResultCode.InternalError
            };

            var errorMessage = context switch
            {
                var c when c.Exception == null => string.Empty,
                var c when _apiConfig.ShowFullException => c.Exception.ToString(),
                _ => context.Exception.Message
            };

            context.ComposedResult = new StandardResult
            {
                StatusCode = statusCode,
                ErrorMessage = errorMessage,
                Data = context.ExecutionResult
            };

            await Task.CompletedTask;
        }

        public virtual async Task OnHttpApiResultExecuted(HttpApiExecutionContext context)
        {
            var content = _resultFormater.Format(context.ComposedResult);
            var buffer = System.Text.Encoding.UTF8.GetBytes(content);
            context.HttpContext.Response.StatusCode = 200;
            await context.HttpContext.Response.Body.WriteAsync(buffer, 0, buffer.Length);
        }


        public virtual async Task OnRpcApiResultExecuting(RpcApiExecutionContext context)
        {
            await Task.CompletedTask;
        }

        public virtual async Task OnRpcApiResultComposition(RpcApiExecutionContext context)
        {
            var statusCode = context switch
            {
                var c when c.Exception == null => ResultCode.OK,
                var c when c.Exception is LogicalException ex => ex.ErrorCode,
                var c when c.Exception is DbRESTFulException ex => ex.ErrorCode,
                _ => ResultCode.InternalError
            };

            var errorMessage = context switch
            {
                var c when c.Exception == null => string.Empty,
                _ => context.Exception.ToString()
            };

            context.ComposedResult = new RpcResponse
            {
                LogicalException = context.Exception != null && context.Exception is LogicalException,
                StatusCode = statusCode,
                ErrorMessage = errorMessage,
                Data = statusCode == ResultCode.OK ? JsonConvert.SerializeObject(context.ExecutionResult) : string.Empty
            };

            await Task.CompletedTask;
        }

        public virtual async Task OnRpcApiResultExecuted(RpcApiExecutionContext context)
        {
            await Task.CompletedTask;
        }
    }
}
