﻿using Microsoft.AspNetCore.Http;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading.Tasks;
using Wicture.DbRESTFul.Resources.Api;

namespace Wicture.DbRESTFul.Infrastructure.Handlers
{
    public class RequestDataParser : IRequestDataParser
    {
        public async Task<JToken> ParseAysnc(ApiParameter apiParameter, HttpApiExecutionContext context)
        {
            var httpContext = context.HttpContext;
            JToken param = null;

            if (apiParameter.query?.Length > 0 || httpContext.Request.Query != null)
            {
                param = ParseQuery(apiParameter.query, httpContext.Request.Query, context.QueryParams);
                VerifyQueryParam(param, apiParameter.query);
                OnQueryDataParsed(param);
            }

            // 当有要求请求体或者有请求体时，进行解析请求体数据
            if (apiParameter.body?.Length > 0 || httpContext.Request.ContentLength > 0)
            {
                var data = await ParseBodyAysnc(httpContext.Request);
                context.BodyContent = data.Item2;
                context.FormParams = data.Item3;

                if (apiParameter.body?.Length > 0)
                {
                    VerifyBodyParam(data.Item1, apiParameter.body, apiParameter.bodyType);
                }
                OnBodyDataParsed(data.Item1);

                param = (param != null && data.Item1 != null && data.Item1 is JObject) ? param.Merge(data.Item1) : data.Item1;
            }

            return param;
        }

        private void VerifyQueryParam(JToken param, Schema[] schema)
        {
            if (param == null) throw new RequestDataParsingException($"The query parameter is required.");
            if (param.Type == JTokenType.Array) throw new RequestDataParsingException("The Array type is not suppored in query parameter.");
            var keys = (param as IDictionary<string, JToken>).Keys.ToArray();
            if (schema == null) return;
            foreach (var item in schema)
            {
                if (item.type == SchemaType.Array || item.type == SchemaType.Object)
                {
                    throw new RequestDataParsingException($"The query param only support simple type, but '{item.name}' type is '{item.type}'.");
                }

                if (!(item.nullable == true) && (!keys.Any(k => k.Equals(item.name, StringComparison.InvariantCultureIgnoreCase)) || param[item.name] == null))
                {
                    throw new RequestDataParsingException($"The parameter '{item.name}' is not nullable.");
                }
            }
        }

        private void VerifyBodyParam(JToken param, Schema[] schema, BodyType bodyType)
        {
          if (param == null) throw new RequestDataParsingException($"The body parameter is required.");
            if (bodyType == BodyType.Array)
            {
                if (param.Type != JTokenType.Array) throw new RequestDataParsingException($"The body parameter must be Array, actually is '{param.Type}'.");

                foreach (var prop in param as JArray)
                {
                    VerifyForSchema(prop, schema);
                }
            }
            else
            {
                if (param.Type != JTokenType.Object) throw new RequestDataParsingException($"The body parameter must be Object, actually is '{param.Type}'.");

                VerifyForSchema(param, schema);
            }
        }

        private void VerifyForSchema(JToken param, Schema[] schema)
        {
            var keys = (param as IDictionary<string, JToken>).Keys.ToArray();
            foreach (var item in schema)
            {
                // It's a definition issue, should not raise exception here, consider to verify it in the UI(PMT)
                //if (item.schema == null && item.type == SchemaType.Object) 
                //{
                //    throw new RequestDataParsingException($"The param '{item.name}' is '{item.type}', schema shuld not be null.");
                //}

                if (item.type == SchemaType.Object && item.schema != null && param[item.name] != null)
                {
                    VerifyForSchema(param[item.name], item.schema);
                }
                else if (item.type == SchemaType.Array && item.schema != null && param[item.name] != null)
                {
                    if (param[item.name].Type != JTokenType.Array)
                    {
                        throw new RequestDataParsingException($"The param '{item.name}' must be an array type, actually is '{param[item.name].Type}'.");
                    }

                    foreach (var prop in param[item.name] as JArray)
                    {
                        VerifyForSchema(prop, item.schema);
                    }
                }
                else if (!(item.nullable == true) && (!keys.Contains(item.name, StringComparison.InvariantCultureIgnoreCase) || param[item.name] == null))
                {
                    throw new RequestDataParsingException($"The parameter '{item.name}' is not nullable.");
                }
            }
        }

        private static JToken ParseQuery(Schema[] schema, IQueryCollection query, Dictionary<string, string> queryParams)
        {
            JToken param = new JObject();
            foreach (var item in query)
            {
                queryParams.TryAdd(item.Key, item.Value.ToString());
                
                var current = schema?.Where(e => e.name.Equals(item.Key, StringComparison.CurrentCultureIgnoreCase)).FirstOrDefault();
                if (current == null)
                {
                    continue;
                }

                if (item.Value.Count == 1)
                {
                    if (current.nullable == true)
                    {
                        if (item.Value.First() == string.Empty)
                        {
                            param[item.Key] = null;
                            continue;
                        }
                    }

                    param[item.Key] = item.Value.First();
                }
                else if (item.Value.Count > 1)
                {
                    param[item.Key] = JToken.FromObject(item.Value.ToArray());
                }
            }

            return param;
        }

        /// <summary>
        /// 解析请求参数
        /// </summary>
        /// <returns>(请求中包含的除文件外的所有参数，除Form类型外的请求体文本内容，Form类型请求体参数)</returns>
        private static async Task<(JToken, string, Dictionary<string, List<string>>)> ParseBodyAysnc(HttpRequest request)
        {
            string contentType = string.Empty;
            if (request.Headers.ContainsKey("Content-Type"))
            {
                contentType = request.Headers["Content-Type"];
            }

            if (!string.IsNullOrEmpty(contentType))
            {
                if (contentType.Contains("multipart/form-data", StringComparison.OrdinalIgnoreCase))
                {
                    request.EnableBuffering();
                    StreamReader reader = new StreamReader(request.Body);
                    var body = await reader.ReadToEndAsync();
                    if (string.IsNullOrEmpty(body)) return (null, string.Empty, null);
                    request.Body.Position = 0;
                    var rst = ParseMultipartFormData(body);
                    // 如果无文件，则直接关闭流
                    if (!rst.Item3)
                    {
                        reader.Close();
                        reader.Dispose();
                    }
                    return (rst.Item1, string.Empty, rst.Item2);
                }
                if (contentType.Contains("application/x-www-form-urlencoded",
                             StringComparison.OrdinalIgnoreCase))
                {
                    using StreamReader reader = new StreamReader(request.Body);
                    var body = await reader.ReadToEndAsync();
                    if (string.IsNullOrEmpty(body)) return (null, body, null);
                    return (ParseBodyForXWWWFormData(body), body, null);
                }
            }

            {
                using StreamReader reader = new StreamReader(request.Body);
                var body = await reader.ReadToEndAsync();
                if (string.IsNullOrEmpty(body)) return (null, body, null);
                return (ParseJsonFormData(body), body, null);
            }
        }

        private static JToken ParseJsonFormData(string body)
        {
            try
            {
                return JToken.Parse(body);
            }
            catch (Exception ex)
            {
                throw new RequestDataParsingException("Failed to parse data from body", ex);
            }
        }

        /// <summary>
        /// 解析Form类型请求参数
        /// </summary>
        /// <returns>(请求体中包含的除文件外的所有参数，请求体中包含的除文件外的所有参数，是否包含文件流)</returns>
        private static (JToken, Dictionary<string, List<string>>, bool) ParseMultipartFormData(string body)
        {
            // Multipart form-data body looks like：
            //------WebKitFormBoundaryUiNRXQW4tEeI03kr
            //Content-Disposition: form-data; name="size"

            //10
            //------WebKitFormBoundaryUiNRXQW4tEeI03kr
            //Content-Disposition: form-data; name="test"

            //asdfasd&asdf?afsdf=
            //------WebKitFormBoundaryUiNRXQW4tEeI03kr--

            using (var stream = new MemoryStream(Encoding.UTF8.GetBytes(body)))
            {
                var parser = new MultipartFormDataParser(stream);
                JToken result = new JObject();

                var kvDic = new Dictionary<string, List<string>>();
                
                foreach (var parameter in parser.Parameters)
                {
                    if (kvDic.ContainsKey(parameter.Name))
                    {
                        if (kvDic[parameter.Name].Count == 1)
                        {
                            result[parameter.Name] = new JArray
                            {
                                result[parameter.Name],
                                parameter.Data
                            };
                        }
                        else
                        {
                            (result[parameter.Name] as JArray).Add(parameter.Data);
                        }
                        kvDic[parameter.Name].Add(parameter.Data);
                    }
                    else
                    {
                        result[parameter.Name] = parameter.Data;
                        kvDic[parameter.Name] = new List<string> { parameter.Data };
                    }
                }

                return (result, kvDic, parser.Files?.Count > 0);
            }
        }

        private static JToken ParseBodyForXWWWFormData(string body)
        {
            JToken result = new JObject();
            string[] content = body.Split('&');
            for (int i = 0; i < content.Length; i++)
            {
                string[] fields = content[i].Split('=');
                if (fields.Length >= 2)
                {
                    result[fields[0]] = WebUtility.UrlDecode(fields[1]);
                }
            }

            return result;
        }

        public virtual void OnQueryDataParsed(JToken param)
        {
            
        }

        public virtual void OnBodyDataParsed(JToken param)
        {
        }
    }
}
