﻿using Microsoft.Extensions.Logging;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using System;
using System.Net;
using System.Security.Cryptography;
using System.Text;
using System.Threading.Tasks;
using Wicture.DbRESTFul.Cache;
using Wicture.DbRESTFul.Infrastructure.Executors;
using Wicture.DbRESTFul.Infrastructure.Filters;

namespace Wicture.DbRESTFul.Infrastructure
{
    public abstract class ApiRquestProcessorBase<CTX> : IApiRequestProcessor<CTX>
         where CTX : ExecutionContext
    {
        protected ILogger _apiLogger;
        protected IFilterInvoker _filterInvoker;
        protected ICacheService _cacheService;
        protected IExecutorProxy _executorProxy;

        public ApiRquestProcessorBase(ILoggerFactory loggerFactory,
            IFilterInvoker filterInvoker,
            ICacheService cacheService,
            IExecutorProxy executorProxy)
        {
            _apiLogger = loggerFactory.CreateLogger("DbRESTFulApi");
            _filterInvoker = filterInvoker;
            _cacheService = cacheService;
            _executorProxy = executorProxy;
        }

        protected virtual async Task OnRequestVerificationAsync(CTX context)
        {
            if (context.Api.deprecated)
            {
                var message = $"The api `{context.Api.name}` is deprecated.";
                throw new DbRESTFulException((int)HttpStatusCode.Gone, message);
            }

            await Task.CompletedTask;
        }

        protected abstract Task OnParameterProcessAsync(CTX context);

        public virtual async Task Process(CTX context)
        {
            try
            {
                await OnRequestVerificationAsync(context);

                await OnParameterProcessAsync(context);

                if (!context.Api.implemented)
                {
                    if (context.Api.mock == null) throw new DbRESTFulException($"No mock data configured for api `{context.Api.name}` in module {context.Api.module}.");

                    await _filterInvoker.InvokeApiMockFilterAsync(context);
                }
                else
                {
                    await _filterInvoker.InvokeAuthorizeFilterAsync(context);

                    await _filterInvoker.InvokeApiExecutingFiltersAsync(context);

                    if (context.CancellationTokenSource.IsCancellationRequested)
                    {
                        throw new DbRESTFulException(ResultCode.LogicError, $"API {context.Api.name} invocation has been canceled.");
                    }
                    else
                    {
                        context.ExecutionResult = await ExcuteApiAsync(context);
                    }

                    await _filterInvoker.InvokeApiExecutedFiltersAsync(context);
                }
            }
            catch (Exception ex)
            {
                context.Exception = ex;
                await _filterInvoker.InvokeApiExceptionFiltersAsync(context);
            }

            await _filterInvoker.InvokeApiResultFilterAsync(context);
        }

        private async Task<object> ExcuteApiAsync(CTX context)
        {
            string cacheKey = null;
            if (context.Api.cache?.enabled == true)
            {
                cacheKey = GetCacheKey(context);
                var cachedResult = await _cacheService.Get(context.Api.cache.type, cacheKey);
                if (cachedResult != null) return cachedResult;
            }

            var result = await _executorProxy.ExecuteAsync(context);

            if (!string.IsNullOrEmpty(cacheKey))
            {
                await _cacheService.Set(context.Api.cache.type, cacheKey, result, context.Api.cache.expiration);
            }

            return result;
        }

        protected virtual string GetCacheKey(CTX context)
        {
            var tokenString = MD5Hash(context.Param);
            return $"{context.Api.implementation.name}_{tokenString}";
        }

        private string MD5Hash(JToken param)
        {
            if (param == null) return string.Empty;
            var input = param.ToString(Formatting.None);

            using (var md5 = MD5.Create())
            {
                var result = md5.ComputeHash(Encoding.ASCII.GetBytes(input));
                return Encoding.ASCII.GetString(result);
            }
        }
    }
}
