﻿using Microsoft.AspNetCore.Http;
using Microsoft.Extensions.Logging;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using System;
using System.Threading.Tasks;
using Wicture.DbRESTFul.Authentication;
using Wicture.DbRESTFul.Infrastructure.Filters;
using Wicture.DbRESTFul.Infrastructure.Handlers;
using Wicture.DbRESTFul.Resources.Api;

namespace Wicture.DbRESTFul.Infrastructure
{
    public class AuthorizationRequestHandler : IRequestHandler
    {      
        private readonly ILogger _logger;
        private readonly TokenOptions _tokenOptions;
        private readonly ITokenExpirationStratagem _tokenExpirationStratagem;
        private readonly IUserIdentifier _userIdentifier;
        private readonly ITokenConverter _tokenConverter;
        private readonly IResultFormater _resultFormater;
        private readonly IRequestDataParser _requestDataParser;
        private readonly IFilterInvoker _filterInvoker;

        public AuthorizationRequestHandler(ILoggerFactory loggerFactory, 
            ITokenOptionsProvider tokenOptionsProvider, 
            ITokenExpirationStratagem tokenExpirationStratagem,
            IUserIdentifier userIdentifier,
            ITokenConverter tokenConverter,
            IResultFormater resultFormater,
            IFilterInvoker filterInvoker,
            IRequestDataParser requestDataParser)
        {
            _logger = loggerFactory.CreateLogger("DbRESTFulApi");
            _tokenOptions = tokenOptionsProvider.Provide();
            _tokenExpirationStratagem = tokenExpirationStratagem;
            _userIdentifier = userIdentifier;
            _tokenConverter = tokenConverter;
            _resultFormater = resultFormater;
            _filterInvoker = filterInvoker;
            _requestDataParser = requestDataParser;
        }

        public int Order { get; set; } = 20;

        public async Task<bool> Handle(HttpApiExecutionContext execContext)
        {
            var context = execContext.HttpContext;
            execContext.Api = new ApiModel
            {
                version = string.Empty,
                owner = string.Empty,
                name = "Token",
                module = "",
                url = context.Request.Path,
                method = HttpMethod.POST,
                title = String.Empty,
                summary = string.Empty,
                note = string.Empty
            };
            if (context.Request.Path.Equals(_tokenOptions.TokenRequestPath, StringComparison.InvariantCultureIgnoreCase))
            {
                await HandleTokenRequestAsync(execContext);
                return true;
            }

            if (context.Request.Path.Equals(_tokenOptions.RefreshTokenRequestPath, StringComparison.InvariantCultureIgnoreCase))
            {
                await HandleRefreshTokenRequestAsync(execContext);
                return true;
            }

            ProcessJwtCookies(context);
            return false;
        }


        private async Task HandleTokenRequestAsync(HttpApiExecutionContext execContext)
        {
            var context = execContext.HttpContext;
            if (!context.Request.Method.Equals("POST"))
            {
                context.Response.StatusCode = 400;
                await context.Response.WriteAsync("Bad request.");
            }
            else
            {
                Exception exception = null;
                try
                {
                    var data = await _requestDataParser.ParseAysnc(new ApiParameter { body = _tokenOptions.LoginSchema }, execContext);
                    execContext.Param = data;

                    _logger.LogDebug($"Token endpoint entry: {context.GetRemoteIpAddressBehindProxy()}, with data: {JsonConvert.SerializeObject(data)}");
                    
                    await _filterInvoker.InvokeApiExecutingFiltersAsync(execContext);
                    
                    var identity = await _userIdentifier.IdentifyAsync(data.Value<string>("username"), data.Value<string>("password"), data, execContext);
                    if (identity == null)
                    {
                        exception = new LogicalException("Invalid username, password or account has been disabled.", 400);
                    }
                    else if (identity.Disabled)
                    {
                        exception = new LogicalException("Account has been disabled.", 403);
                    }
                    else
                    {
                        JObject token = _tokenConverter.Compose(identity);

                        _logger.LogDebug($"Token generated: {token}.");
                        _logger.LogInformation($"User Id: {identity.Id}, Name: {identity.Name} has logged in.");
                        SetJwtCookies(context, identity, token);

                        execContext.ExecutionResult = token;
                    }
                }
                catch(Exception ex)
                {
                    _logger.LogError(ex, "Token generation error");
                    exception = ex;
                }

                if (exception != null) execContext.Exception = exception;
                await _filterInvoker.InvokeApiExecutedFiltersAsync(execContext);
                await _filterInvoker.InvokeApiResultFilterAsync(execContext);
            }
        }

        public async Task HandleRefreshTokenRequestAsync(HttpApiExecutionContext execContext)
        {
            var context = execContext.HttpContext;
            execContext.Api = new ApiModel
            {
                version = string.Empty,
                owner = string.Empty,
                name = "Token",
                module = "",
                url = context.Request.Path,
                method = HttpMethod.POST,
                title = String.Empty,
                summary = string.Empty,
                note = string.Empty
            };
            if (!context.Request.Method.Equals("POST"))
            {
                context.Response.StatusCode = 400;
                await context.Response.WriteAsync("Bad request.");
            }
            else
            {
                Exception exception = null;
                try
                {
                    var token = context.Request.Headers["Authorization"].ToString();
                    if (token.Length < TokenOptions.StartIndex)
                    {
                        exception = new Exception("The length of token is invalid.");
                    }
                    else
                    {
                        await _filterInvoker.InvokeApiExecutingFiltersAsync(execContext);

                        UserIdentity identity = _tokenConverter.Parse(token.Substring(TokenOptions.StartIndex));
                        JObject userToken = _tokenConverter.Compose(identity);

                        _logger.LogDebug($"token generated: {token}.");

                        execContext.ExecutionResult = userToken;
                    }
                }
                catch (Exception ex)
                {
                    _logger.LogDebug(ex, "Unauthorized, Token validation failed.");
                    exception = new LogicalException("Token invalid.", 403);
                }
                if (exception != null) execContext.Exception = exception;
                await _filterInvoker.InvokeApiExecutedFiltersAsync(execContext);
                await _filterInvoker.InvokeApiResultFilterAsync(execContext);
            }
        }

        private void ProcessJwtCookies(HttpContext context)
        {
            if (_tokenOptions.UseJwtCookie && context.Request.Cookies.TryGetValue(_tokenOptions.CookieName, out var accessToken))
            {
                if (_tokenOptions.ForceCookie)
                {
                    if (string.IsNullOrEmpty(accessToken))
                    {
                        context.Request.Headers.Remove("Authorization");
                    }
                }
                if (context.Request.Headers.TryGetValue("Authorization", out var auth))
                {
                    if (string.IsNullOrEmpty(auth))
                    {
                        if (!string.IsNullOrEmpty(accessToken))
                        {
                            string bearerToken = $"bearer {accessToken}";
                            context.Request.Headers.Add("Authorization", bearerToken);
                        }
                    }
                }
            }
        }

        private void SetJwtCookies(HttpContext context, UserIdentity identity, JObject token)
        {
            if (!_tokenOptions.UseJwtCookie) return;

            context.Response.Cookies.Append(
                _tokenOptions.CookieName,
                token.Value<string>(_tokenOptions.CookieName),
                new CookieOptions()
                {
                    HttpOnly = true,
                    Expires = DateTime.Now.Add(_tokenExpirationStratagem.GetExpirationSpan(_tokenOptions, identity))
                });
        }
    }
}
