﻿using Microsoft.AspNetCore.Http;
using Newtonsoft.Json;
using System.Linq;
using System.Threading.Tasks;
using Wicture.DbRESTFul.Configuration;
using Wicture.DbRESTFul.DevTool;
using Wicture.DbRESTFul.Resources;

namespace Wicture.DbRESTFul.Infrastructure
{
    public class DevToolRequestHandler : IRequestHandler
    {
        private bool syncOnloading = false;
        private const string ApplicationTitle = "###APPLICATION###";
        private const string TemplateKeywords = "###SYNCLINKS###";

        private readonly ResourceSynchronizer _synchronizer;
        private readonly DevToolSectionConfig _devToolConfig;

        private readonly IResourceManager _resourceManager;

        public DevToolRequestHandler(ISectionConfigProvider sectionConfigProvider, IResourceManager resourceManager)
        {
            _devToolConfig = sectionConfigProvider.Provide<DevToolSectionConfig>();
            if (_devToolConfig.Enabled)
            {
                _synchronizer = new ResourceSynchronizer(sectionConfigProvider, resourceManager);
            }

            _resourceManager = resourceManager;
        }

        private async Task RenderDevPage(HttpApiExecutionContext execContext)
        {
            var context = execContext.HttpContext;
            var data = await _synchronizer.ListModulesAsync();
            context.Response.Headers["Content-Type"] = "text/html;charset=utf-8";

            var projectId = data.FirstOrDefault()?.ProjectId;

            if (projectId == null || projectId < 1)
            {
                await context.Response.WriteAsync($"ProjectName `{_devToolConfig?.ProjectName}` is not valid, please check config.");
                return;
            }

            var list = $"<ul class='sync-links'><li><a data-url='/dev/api/sync?projectId={projectId}' href='#'>同步所有Api</a></li>";
            foreach (var item in data)
            {
                list += $"<li><a data-url='/dev/api/sync?projectId={projectId}&moduleId={item.Id}' href='#'>同步模块{item.Name}</a></li>";
            }
            list += "</ul>";

            list += $"<hr />";
            list += $"<ul class='sync-links'><li><a data-url='/dev/csi/sync?projectId={projectId}' href='#'>同步所有Csi</a></li>";
            foreach (var item in data)
            {
                list += $"<li><a data-url='/dev/csi/sync?projectId={projectId}&moduleId={item.Id}' href='#'>同步模块{item.Name}</a></li>";
            }
            list += "</ul>";

            var html = Resource.DevToolPageHtml.Replace(TemplateKeywords, list).Replace(ApplicationTitle, _devToolConfig?.ProjectName);

            await context.Response.WriteAsync(html);

            if (_devToolConfig?.SyncOnLoading == true && !syncOnloading)
            {
                _synchronizer.SyncApisAsync(projectId.ToString(), null).Wait();
                _synchronizer.SyncCsisAsync(projectId.ToString(), null).Wait();

                syncOnloading = true;
            }
        }

        public int Order { get; set; } = 40;

        public async Task<bool> Handle(HttpApiExecutionContext execContext)
        {
            if (!_devToolConfig.Enabled) return false;

            var context = execContext.HttpContext;
            bool processed = false;
            if (context.Request.Path.Value.Equals(string.IsNullOrEmpty(_devToolConfig.Path) ? "/dev" : _devToolConfig.Path))
            {
                await RenderDevPage(execContext);
                processed = true;
            }
            else if (context.Request.Path == "/dev/api/sync")
            {
                var projectId = context.Request.Query["projectId"].FirstOrDefault();
                var moduleId = context.Request.Query["moduleId"].FirstOrDefault();
                var data = await _synchronizer.SyncApisAsync(projectId, moduleId);
                var content = JsonConvert.SerializeObject(data, Formatting.Indented);
                await context.Response.WriteAsync(content);
                processed = true;
            }
            else if (context.Request.Path == "/dev/csi/sync")
            {
                var projectId = context.Request.Query["projectId"].FirstOrDefault();
                var moduleId = context.Request.Query["moduleId"].FirstOrDefault();
                var data = await _synchronizer.SyncCsisAsync(projectId, moduleId);
                var content = JsonConvert.SerializeObject(data, Formatting.Indented);
                await context.Response.WriteAsync(content);
                processed = true;
            }

            if (processed)
            {
                _resourceManager.LoadAll();
            }

            return processed;
        }
    }
}
