﻿using Dapper;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using System.Threading.Tasks;
using Wicture.DbRESTFul.Cache;
using Wicture.DbRESTFul.Infrastructure.Filters;
using Wicture.DbRESTFul.Infrastructure.Repository.Middlewares;
using Wicture.DbRESTFul.Resources;
using Wicture.DbRESTFul.Resources.Csi;

namespace Wicture.DbRESTFul.Infrastructure.Repository
{
    public class DbRESTFulRepository : IDbRESTFulRepository, IAsyncDisposable, IDisposable
    {
        public IFilterInvoker FilterInvoker { get; set; }

        public ExecutionContext Context { get; set; }

        public IConnectionManager ConnectionManager { get; set; }

        public IResourceManager ResourceManager { get; set; }

        private InvokeContext currentContext;

        public async Task<dynamic> InvokeAsync(string csiName, object parameters = null, IDbConnection connection = null, IDbTransaction transaction = null)
        {
            var csi = GetCsiWithName(csiName);
            var context = await PrepareContextAsync(csi, parameters, connection, transaction);

            try
            {
                if (csi.IsSingleResultSet)
                {
                    await InvokeSingleCode<dynamic>(context);
                }
                else
                {
                    await InvokeMultiCode<dynamic>(context);
                }

                if (context.AutoManagedTransaction) context.Transaction?.Commit();

                return context.Result;
            }
            catch (Exception ex)
            {
                if (context.AutoManagedTransaction) context.Transaction?.Rollback();

                throw new CsiInvocationException($"Invoke {csiName} failed with parameters: {Environment.NewLine}{JsonConvert.SerializeObject(parameters)}", ex);
            }
            finally
            {
                if (context.AutoManagedConnection)
                {
                    context.Connection.Close();
                    context.Connection.Dispose();
                }
            }
        }

        public async Task<IEnumerable<T>> InvokeAsync<T>(string csiName, object parameters = null, IDbConnection connection = null, IDbTransaction transaction = null)
        {
            var context = await InvokeInternalAsync<T>(csiName, parameters, connection, transaction);
            return context.Result as IEnumerable<T>;
        }

        public async Task<T> InvokeSingleOrDefaultAsync<T>(string csiName, object parameters = null, IDbConnection connection = null, IDbTransaction transaction = null)
        {
            var context = await InvokeInternalAsync<T>(csiName, parameters, connection, transaction);
            return (T)context.Result;
        }

        public async Task<IEnumerable<T>> QueryAsync<T>(string sql, object parameters = null, IDbConnection connection = null, IDbTransaction transaction = null, bool queryOnly = true)
        {
            var csi = CsiModel.Build(sql, false, queryOnly, transaction != null);
            var context = await QueryInternalAsync<T>(csi, parameters, connection, transaction);
            return context.Result as IEnumerable<T>;
        }

        public async Task<T> QuerySingleOrDefaultAsync<T>(string sql, object parameters = null, IDbConnection connection = null, IDbTransaction transaction = null, bool queryOnly = true)
        {
            var csi = CsiModel.Build(sql, true, queryOnly, transaction != null);
            var context = await QueryInternalAsync<T>(csi, parameters, connection, transaction);
            return (T)context.Result;
        }

        private async Task<InvokeContext> InvokeInternalAsync<T>(string invokeName, object parameters, IDbConnection connection = null, IDbTransaction transaction = null)
        {
            var csi = GetCsiWithName(invokeName);
            csi.EnsureSingleResultSet();

            var context = await PrepareContextAsync(csi, parameters, connection, transaction);

            try
            {
                await InvokeSingleCode<T>(context);
                if (context.AutoManagedTransaction) context.Transaction?.Commit();

                return context;
            }
            catch (Exception ex)
            {
                if (context.AutoManagedTransaction) context.Transaction?.Rollback();

                throw new CsiInvocationException($"Invoke {invokeName} failed with parameters: {Environment.NewLine}{JsonConvert.SerializeObject(parameters)}", ex);
            }
            finally
            {
                if (context.AutoManagedConnection)
                {
                    context.Connection.Close();
                    context.Connection.Dispose();
                }
            }
        }

        private async Task<InvokeContext> PrepareContextAsync(CsiModel csi, object parameters, IDbConnection connection, IDbTransaction transaction)
        {
            var conn = connection ?? ConnectionManager.GetConnection(csi.queryOnly);

            if (conn.State != ConnectionState.Open)
            {
                conn.Open();
            }

            var context = new InvokeContext
            {
                Repository = this,
                Csi = csi,
                Connection = conn,
                Transaction = transaction ?? (csi.requiredTransaction ? conn.BeginTransaction() : null),

                // if connection specified, don't dispose, i.e. should manage in the caller scope.
                AutoManagedConnection = connection == null,

                // if transaction specified, don't handle commit and rollback.
                AutoManagedTransaction = transaction == null
            };

            context.Param = ParametersProcessor.WrapParameters(parameters ?? new JObject());

            await FilterInvoker.ProcessCsiMiddlewareAsync(context, ActivePosition.BeforeBuildCommandParam);
            context.CommandParam = ParametersProcessor.BuildDynamicParameters(context.Param);

            await FilterInvoker.InvokeCsiInvokingFiltersAsync(context);

            await FilterInvoker.ProcessCsiMiddlewareAsync(context, ActivePosition.BeforeInvoke);

            currentContext = context;

            return context;
        }

        private async Task InvokeSingleCode<T>(InvokeContext context)
        {
            var resultSetInfo = context.Csi.ListResultSetInfo().FirstOrDefault();
            var items = await RunSql<T>(resultSetInfo.Sql, context);
            context.Result = resultSetInfo.MultipleRecords ? items : (object)items.FirstOrDefault();

            await FilterInvoker.ProcessCsiMiddlewareAsync(context, ActivePosition.AfterInvoke);
            await FilterInvoker.InvokeCsiInvokedFiltersAsync(context);
        }

        private async Task InvokeMultiCode<T>(InvokeContext context)
        {
            var resultSetInfos = context.Csi.ListResultSetInfo();
            var result = new JObject();
            foreach (var info in resultSetInfos)
            {
                var items = await RunSql<dynamic>(info.Sql, context);
                var data = info.MultipleRecords ? items : (object)items.FirstOrDefault();
                result.Add(info.Key, data == null ? null : JToken.FromObject(data));
            }

            context.Result = result;

            await FilterInvoker.ProcessCsiMiddlewareAsync(context, ActivePosition.AfterInvoke);

            await FilterInvoker.InvokeCsiInvokedFiltersAsync(context);
        }

        private async Task<InvokeContext> QueryInternalAsync<T>(CsiModel csi, object parameters, IDbConnection connection, IDbTransaction transaction)
        {
            var context = await PrepareContextAsync(csi, parameters, connection, transaction);
            try
            {
                await InvokeSingleCode<T>(context);
                if (context.AutoManagedTransaction) context.Transaction?.Commit();
                return context;
            }
            catch (Exception ex)
            {
                if (context.AutoManagedTransaction) context.Transaction?.Rollback();
                throw new CsiInvocationException($"Query sql `{csi.code}` failed with parameters: {Environment.NewLine}{JsonConvert.SerializeObject(parameters)}", ex);
            }
            finally
            {
                if (context.AutoManagedConnection)
                {
                    context.Connection.Close();
                    context.Connection.Dispose();
                }
            }
        }

        private async Task<IEnumerable<T>> RunSql<T>(string sqlCode, InvokeContext context)
        {
            string sql = ParametersProcessor.ProcessOptionalParameters(sqlCode, context.CommandParam);

            try
            {
                // Array Parameters is mainly used for insertion and update,
                // Which can only call the SqlMapper.Execute.
                IEnumerable<T> result;
                if (context.CommandParam is List<DynamicParameters>)
                {
                    int efforts = await SqlMapper.ExecuteAsync(context.Connection, sql, context.CommandParam, context.Transaction, context.Repository.ConnectionManager.DbConnection?.CommandTimeout);

                    T data;
                    try
                    {
                        data = (T)Convert.ChangeType(efforts, typeof(T));
                        result = new List<T> { data };
                    }
                    catch
                    {
                        throw new DbRESTFulException("Array parameters only support build-in primary return type: (int, decimal, double, string. etc.)");
                    }
                }
                else
                {
                    result = await SqlMapper.QueryAsync<T>(context.Connection, sql, context.CommandParam, context.Transaction, context.Repository.ConnectionManager.DbConnection?.CommandTimeout);
                }

                return result;
            }
            catch (Exception ex)
            {
                throw new DbRESTFulException($"Invoke code failed: {sqlCode}, " +
                    $"{Environment.NewLine}with parameters: {JsonConvert.SerializeObject(context.CommandParam)}", ex);
            }
        }

        private CsiModel GetCsiWithName(string invokeName)
        {
            if (!ResourceManager.Csis.TryGetValue(invokeName, out CsiModel csi))
            {
                throw new DbRESTFulException($"The csi `{invokeName}` is not found with name.");
            }

            return CsiModel.Clone(csi);
        }

        private bool disposed = false;

        public ValueTask DisposeAsync()
        {
            try
            {
                Dispose();
                return default;
            }
            catch (Exception exception)
            {
                return new ValueTask(Task.FromException(exception));
            }
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (disposed) return;

            if (currentContext?.Connection?.State == ConnectionState.Open)
            {
                try
                {
                    currentContext.Connection.Close();
                    currentContext.Connection.Dispose();
                    currentContext = null;
                }
                catch (Exception)
                {
                    // ignore.
                }
            }

            disposed = true;
        }
    }
}