﻿using Newtonsoft.Json.Linq;
using System;
using System.Linq;
using System.Threading.Tasks;
using Wicture.DbRESTFul.Resources.Csi;

namespace Wicture.DbRESTFul.Infrastructure.Repository.Middlewares
{
    /// <summary>
    /// 分页中间件，对查询结果做自动分页支持。
    /// </summary>
    public class PaginationMiddleware : ICsiMiddleware
    {
        public string Name { get { return "pagination"; } }
        public ActivePosition ActivePosition { get { return ActivePosition.AfterInvoke; } }

        public string Validate(InvokeContext context, JToken config)
        {
            var message = string.Empty;
            if (!(config is JObject))
            {
                message = "config type should be an object.";
            }

            if (!(context.Csi.code is JValue))
            {
                var prefix = string.IsNullOrEmpty(message) ? string.Empty : Environment.NewLine;
                message += prefix + "pagination middle ware only support single query code.";
            }

            if (!(context.Param is JObject))
            {
                var prefix = string.IsNullOrEmpty(message) ? string.Empty : Environment.NewLine;
                message += prefix + "parameter type should be an object.";
            }
            else
            {
                var setting = context.Param as JObject;
                if (string.IsNullOrEmpty(setting.Value<string>("pageSize"))
                    || string.IsNullOrEmpty(setting.Value<string>("pageIndex"))
                    || string.IsNullOrEmpty(setting.Value<string>("pageStart")))
                {
                    var prefix = string.IsNullOrEmpty(message) ? string.Empty : Environment.NewLine;
                    message = prefix + "pageSize, pageIndex or pageStart is not set.";
                }
            }

            return message;
        }

        /// <summary>
        /// Code could be:
        ///     "SELECT u.`name`, u.score FROM user AS u WHERE u.id > @userId limit @pageStart, @pageSize",
        /// Need to convert:
        ///     "SELECT count(u.`name`) as count, @pageSize as size, @pageIndex as `index` FROM user AS u WHERE u.id > @userId"
        /// </summary>
        /// <param name="context"></param>
        /// <param name="config">
        /// {
        ///     "size": "pageSize",
        ///     "count": "totalCount",
        ///     "page": "pageIndex"
        /// }
        /// </param>
        public async Task ProcessAsync(InvokeContext context, JToken config)
        {
            var setting = config as JObject;
            var countPropertyName = string.IsNullOrEmpty(setting.Value<string>("count")) ? "count" : setting.Value<string>("count");
            var sizePropertyName = string.IsNullOrEmpty(setting.Value<string>("size")) ? "size" : setting.Value<string>("size");
            var pagePropertyName = string.IsNullOrEmpty(setting.Value<string>("page")) ? "page" : setting.Value<string>("page");

            var paginationSql = ConvertPaginationCode(context, countPropertyName, sizePropertyName, pagePropertyName);
            var pagination = await context.Repository.QuerySingleOrDefaultAsync<dynamic>(paginationSql, context.Param, context.Connection);

            context.Result = new { items = context.Result, pagination };
        }

        /// <summary>
        /// Convert query code to pagination code.
        /// </summary>
        /// <param name="code">
        /// Code could be like:
        ///     "SELECT u.`name`, u.score FROM user AS u WHERE u.id > @userId ORDER  BY u.id LIMIT @pageStart, @pageSize",
        /// </param>
        /// <returns>
        /// Converted code could be like:
        ///     "SELECT count(u.`name`) as count, @pageSize as size, @pageIndex as `index` FROM user AS u WHERE u.id > @userId"
        /// </returns>
        private string ConvertPaginationCode(InvokeContext context, string countPropertyName, string sizePropertyName, string pagePropertyName)
        {
            var code = context.Csi.code.ToString();
            if (context.Repository.ConnectionManager.DbConnection.DatabaseType == DatabaseType.MySQL)
            {
                // Get code copy for `FROM ... WHERE ....`,
                var fromAndWhere = GetFromAndWhereForMySQL(code);

                return $"SELECT count(1) AS `{countPropertyName}`, @pageSize AS `{sizePropertyName}`, @pageIndex AS `{pagePropertyName}` {fromAndWhere}";
            }
            else
            {
                var fromAndWhere = GetFromAndWhereForSQLServer(code);
                return $"SELECT count(1) AS {countPropertyName}, @pageSize AS {sizePropertyName}, @pageIndex AS {pagePropertyName} {fromAndWhere} ";
            }
        }

        /// <summary>
        /// 去除字符串头尾的圆括号对
        /// </summary>
        private string TrimParentheses(string str)
        {
            if (string.IsNullOrEmpty(str))
                return str;
            if (str.StartsWith("(") && str.EndsWith(")"))
                return TrimParentheses(str.Substring(1, str.Length - 2));
            else
                return str;
        }

        /// <summary>
        /// 判断在某位置紧接的前面是否有奇数的反斜杠，如\\\'，在'前就有奇数的\
        /// </summary>
        private bool HasOddBackslashBefore(string str, int index)
        {
            var count = 0;
            for (var i = index - 1; i > -1; i--)
            {
                if (str[i] == '\\')
                {
                    count++;
                }
                else
                    break;
            }
            return count % 2 == 1;
        }

        /// <summary>
        /// 获取主查询语句片断，去除子查询及字符串。去除之前执行了Trim()
        /// </summary>
        private string GetValidMainSqlClip(string str)
        {
            str = TrimParentheses(str.Trim());
            var validChars = new char[str.Length];
            var quote = ' ';
            var parenthesesCount = 0;
            for (var i = 0; i < str.Length; i++)
            {
                validChars[i] = ' ';
                if ((str[i] == '\"' || str[i] == '\'') && !HasOddBackslashBefore(str, i))
                {
                    quote = quote == ' ' ? str[i] : ' ';
                    continue;
                }
                if (quote != ' ') continue;
                if (str[i] == '(')
                    parenthesesCount++;
                else if (str[i] == ')')
                    parenthesesCount--;
                else if(parenthesesCount == 0)
                    validChars[i] = str[i];
            }
            return new string(validChars);
        }

        private string GetFromAndWhereForMySQL(string code)
        {
            try
            {
                string str = code.Trim();

                string validMainSqlClip = GetValidMainSqlClip(str).ToUpperInvariant();

                var keywords = "ORDER,LIMIT";
                var positions = keywords.Split(',').Select(kw => validMainSqlClip.IndexOf(" " + kw + " ")).Where(i => i > 0);
                var minIndex = str.Length;
                if (positions.Count() > 0)
                {
                    minIndex = positions.Min();
                }
                var startIndex = validMainSqlClip.IndexOf(" FROM ");

                return str.Substring(startIndex, minIndex - startIndex).Trim();
            }
            catch (Exception e)
            {
                throw new Exception("Auto Pagination failed, please check the csi code.", e);
            }
        }

        private string GetFromAndWhereForSQLServer(string code)
        {
            try
            {
                string str = code.Trim();

                string validMainSqlClip = GetValidMainSqlClip(str).ToUpperInvariant();

                var keywords = "ORDER,OFFSET";
                var positions = keywords.Split(',').Select(kw => validMainSqlClip.IndexOf(" " + kw + " ")).Where(i => i > 0);
                var minIndex = str.Length;
                if (positions.Count() > 0)
                {
                    minIndex = positions.Min();
                }
                var startIndex = validMainSqlClip.IndexOf(" FROM ");

                return str.Substring(startIndex, minIndex - startIndex).Trim();
            }
            catch (Exception e)
            {
                throw new Exception("Auto Pagination failed, please check the csi code.", e);
            }
        }
    }
}
