﻿using Dapper;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.Dynamic;
using System.Linq;
using System.Text.RegularExpressions;

namespace Wicture.DbRESTFul.Infrastructure.Repository
{
    public static class ParametersProcessor
    {
        private static readonly Regex reg = new Regex(@"\[([^\[\]@]*@[^\[\]@]*)*\]");
        private static readonly Regex pReg = new Regex(@"\@[a-zA-Z_]+\w*");

        public static object BuildDynamicParameters(JToken param)
        {
            if (param == null) return null;

            if (param is JObject)
            {
                return new DynamicParameters(BuildParamForJObject(param as JObject));
            }
            else if(param is JArray)
            {
                var result = new List<DynamicParameters>();
                foreach (var item in param as JArray)
                {
                    result.Add(new DynamicParameters(BuildParamForJObject(item as JObject)));
                }

                return result;
            }

            throw new Exception("Unsupported param type: {0}".FormatWith(param.GetType()));
        }

        public static string ProcessOptionalParameters(string code, object commandParam)
        {
            var matches = reg.Matches(code);
            var sql = code;
            foreach (var match in matches)
            {
                var str = match.ToString();
                var containAll = true;

                if (commandParam != null)
                {
                    // Get first CommandParam if it's an array.
                    var param = commandParam is List<DynamicParameters>
                        ? (commandParam as List<DynamicParameters>).FirstOrDefault()
                        : commandParam as DynamicParameters;

                    foreach (var subMath in pReg.Matches(str))
                    {
                        if (!containAll)
                            break;
                        var exist = false;

                        foreach (var p in param.ParameterNames)
                        {
                            // Skip the null param.
                            if (subMath.ToString() == $"@{p}" && param.Get<dynamic>(p) != null)
                            {
                                exist = true;
                                break;
                            }
                        }

                        containAll = containAll && exist;
                    }
                }
                else
                {
                    containAll = false;
                }

                sql = containAll
                    ? sql.Replace(str, str.Replace("[", "").Replace("]", ""))
                    : sql.Replace(str, "");
            }

            return sql;
        }

        /// <summary>
        /// Wrap parameters in to JToken.
        /// </summary>
        public static JToken WrapParameters(object parameters)
        {
            JToken token = null;

            try
            {
                if (parameters != null)
                {
                    token = parameters is JToken
                        ? parameters as JToken
                        : (parameters is string) ? JToken.Parse(parameters.ToString()) : JToken.FromObject(parameters);
                }

                ProcessPaginationParameters(token);

                return token;
            }
            catch (Exception ex)
            {
                if (ex is ArgumentException)
                {
                    throw;
                }

                throw new ArgumentException($"The expected parameters should be JObject, JArrary and string (json format).", ex);
            }
        }

        /// <summary>
        /// TODO: Pagination could process in the PaginationMiddleware.
        /// </summary>
        /// <param name="param"></param>
        private static void ProcessPaginationParameters(JToken param)
        {
            if (param == null || !(param is JObject)) return;

            var jObj = param as JObject;
            try
            {
                var pageSize = jObj.Value<int>("pageSize");
                var pageIndex = jObj.Value<int>("pageIndex");
                if (pageSize > 0 && pageIndex == 0)
                {
                    throw new Exception("'pageIndex' must be greater than 0");
                }
                // Only if the pageSize and pageIndex parameters set with validated values.
                else if (pageSize > 0 && pageIndex > 0)
                {
                    jObj["pageSize"] = pageSize;    // Make sure it is int type.
                    jObj["pageIndex"] = pageIndex;  // Make sure it is int type.
                    jObj["pageStart"] = pageSize * (pageIndex - 1);
                }
            }
            catch (Exception ex)
            {
                throw new ArgumentException("Parameter for pagination is incorrect.", ex);
            }
        }

        private static object BuildParamForJObject(JObject param)
        { 
            var data = new ExpandoObject() as IDictionary<string, object>;

            foreach (var item in param)
            {
                object val = null;

                if (item.Value is JArray)
                {
                    val = (item.Value as JArray).Values().Select(v => v.ToString());
                }
                else
                {
                    val = (item.Value as JValue).Value;
                }

                data.Add(item.Key, val);
            }

            return data;
        }
    }
}