﻿using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using Newtonsoft.Json.Serialization;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Wicture.DbRESTFul.Configuration;

namespace Wicture.DbRESTFul.Resources.Csi
{
    public class CsiLoader : IResourceLoader<CsiModel>
    {
        private static string[] WriteOperationKeywords = new[] { "Insert ", "Update ", "Delete " };
        private const string csiFileExtension = "*.*";

        public event EventHandler Loaded;

        public Dictionary<string, CsiModel> Items { get; } = new Dictionary<string, CsiModel>();

        private readonly CsiSectionConfig _csiConfig;

        public CsiLoader(ISectionConfigProvider sectionConfigProvider)
        {
            _csiConfig = sectionConfigProvider.Provide<CsiSectionConfig>();
        }

        public void Load()
        {
            Items.Clear();

            if (string.IsNullOrEmpty(_csiConfig.Path)) throw new ResourceLoadingException("Csi path is not specified");

            var targetDir = Path.Combine(Directory.GetCurrentDirectory(), _csiConfig.Path);

            if (!Directory.Exists(targetDir)) Directory.CreateDirectory(targetDir);

            string[] files = Directory.GetFiles(targetDir, csiFileExtension);

            string currentFile = string.Empty;
            try
            {
                foreach (var file in files)
                {
                    currentFile = file;
                    string jsonData = File.ReadAllText(file);

                    var items = JsonConvert.DeserializeObject<List<CsiModel>>(jsonData);

                    foreach (var item in items)
                    {
                        if (Items.ContainsKey(item.name)) throw new ResourceLoadingException($"Csi name {item.name} already exists.");

                        if (item.queryOnly) VerifyQueryOnly(item);

                        Items.Add(item.name, item);
                    }
                }

                Loaded?.Invoke(this, null);
            }
            catch (Exception ex)
            {
                throw new ResourceLoadingException($"Parsing Csi file:'{currentFile}' failed.", ex);
            }
        }

        public class CaseInsensitiveContractResolver : DefaultContractResolver
        {
            protected override string ResolvePropertyName(string propertyName)
            {
                if (propertyName == "middleWares")
                {
                    propertyName = "middlewares";
                }

                return base.ResolvePropertyName(propertyName);
            }
        }

        /// <summary>
        /// Verify if the queryOnly csi has Update, Insert or Delete operation (sql keyword).
        /// </summary>
        private void VerifyQueryOnly(CsiModel item)
        {
            var validate = new Func<object, bool>(code => !WriteOperationKeywords.Any(keyword => code.ToString().Contains(keyword, StringComparison.CurrentCultureIgnoreCase)));

            if (item.code.Type == JTokenType.String && !validate(item.code.Value<string>()))
            {
                throw new ResourceLoadingException($"Code for QueryOnly Csi `{item.name}` has Write Operation Keywords.");
            }

            foreach (var code in item.code.OfType<JObject>())
            {
                if (!validate(code.Value<string>()))
                {
                    throw new ResourceLoadingException($"Code for QueryOnly Csi `{item.name}` has Write Operation Keywords.");
                }
            }
        }
    }
}
