﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Runtime.Loader;
using Wicture.DbRESTFul.Configuration;
using Wicture.DbRESTFul.Infrastructure.Repository;

namespace Wicture.DbRESTFul.Resources.Repository
{
    public class RepositoryLoader : IResourceLoader<RepositoryModel>
    {
        public Dictionary<string, RepositoryModel> Items { get; } = new Dictionary<string, RepositoryModel>();

        public event EventHandler Loaded;

        public List<Assembly> Assemblies { get; private set; } = new List<Assembly>();


        private readonly RepositorySectionConfig _repositoryConfig;

        public RepositoryLoader(ISectionConfigProvider sectionConfigProvider)
        {
            _repositoryConfig = sectionConfigProvider.Provide<RepositorySectionConfig>();
        }

        public void Load()
        {
            Items.Clear();

            try
            {
                LoadAssemblies(_repositoryConfig);
                var baseType = typeof(IDbRESTFulRepository);
                foreach (var assembly in Assemblies)
                {
                    assembly.GetTypes()
                        .Where(t => baseType.IsAssignableFrom(t) && t.Name != baseType.Name && !t.GetTypeInfo().IsAbstract && !t.GetTypeInfo().IsInterface)
                        .ForEach(t => LoadMethods(t));
                }

                var dependences = new List<string>(_repositoryConfig.Dependences?? new string[] { });

                foreach (var item in dependences)
                {
                    var dll = item;
                    if (!item.EndsWith(".dll", StringComparison.InvariantCultureIgnoreCase))
                    {
                        dll += ".dll";
                    }

                    var path = Path.Combine(_repositoryConfig.Path, dll);
                    if (!File.Exists(path)) path = Path.Combine(Directory.GetCurrentDirectory(), dll);
                    if (!File.Exists(path)) path = Path.Combine(Path.GetDirectoryName(Assembly.GetEntryAssembly().Location), dll);

                    if (!File.Exists(path))
                    {
                        throw new FileNotFoundException($"dependency file is not found `{path}`");
                    }

                    AssemblyLoadContext.Default.LoadFromAssemblyPath(path);
                }

                Loaded?.Invoke(this, null);
            }
            catch (Exception ex)
            {
                LoggerManager.Logger.Error(ex, "Failed to load Repositories.");
                throw ex;
            }
        }

        private void LoadAssemblies(RepositorySectionConfig setting)
        {
            Assemblies.Clear();

            if(Directory.Exists(setting.Path))
            {
                var dllfiles = Directory.GetFiles(setting.Path, "*.dll")
                    .Where(n =>
                    {
                        var fn = Path.GetFileNameWithoutExtension(n);
                        return !fn.Equals(Path.GetFileNameWithoutExtension(Assembly.GetEntryAssembly().Location))
                            && setting.Includes?.Select(f => f.ToLower().TrimEnd(".dll".ToCharArray())).Contains(fn.ToLower()) == true;
                    });

                Assemblies.AddRange(dllfiles.Select(file => AssemblyLoadContext.Default.LoadFromAssemblyPath(file)));
            }
            Assemblies.Add(Assembly.GetEntryAssembly());

            if (setting.Includes != null)
            {
                foreach (var item in setting.Includes)
                {
                    if (!Assemblies.Any(a => a.Location.Contains(item, StringComparison.CurrentCultureIgnoreCase)))
                    {
                        throw new Exception($"Assemblies: {item} is not found, which configured in includes.");
                    }
                }
            }

            LoggerManager.Logger.Info($"Loaded assemblies: {string.Join(Environment.NewLine, Assemblies.Select(a => a.FullName))} in RepositoryLoader");
        }

        private void LoadMethods(Type repositoryType)
        {
            try
            {
                foreach (var item in repositoryType.GetMethods(BindingFlags.Instance | BindingFlags.Public | BindingFlags.DeclaredOnly))
                {
                    Items.Add($"{repositoryType.Name}.{item.Name}", new RepositoryModel(repositoryType, item));
                }
            }
            catch (Exception ex)
            {
                throw new Exception($"Create instance for Repository type {repositoryType} failed, please make sure it has default constructor.", ex);
            }
        }
    }
}
