﻿using Grpc.Core;
using Microsoft.Extensions.Logging;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using Polly;
using System;
using System.Diagnostics;
using System.Linq;
using System.Threading.Tasks;

namespace Wicture.DbRESTFul.Rpc.Client
{
    public class RpcClient : IRpcClient
    {
        private readonly IRpcServiceRouter _rpcServiceRouter;
        private readonly ILogger _logger;
        private readonly IGrpcChannelProvider _grpcChannelProvider;

        public RpcClient(IRpcServiceRouter rpcServiceRouter, IGrpcChannelProvider grpcChannelProvider, ILoggerFactory loggerFactory)
        {
            _rpcServiceRouter = rpcServiceRouter;
            _grpcChannelProvider = grpcChannelProvider;
            _logger = loggerFactory.CreateLogger("DbRESTFulApi");
        }

        public async Task<RpcResponse> InvokeRpcAsync(ExecutionContext executionContext)
        {
            var route = _rpcServiceRouter.Resolve(executionContext);
            return await InvokeAsync(route.ServiceName, route.ApiName, executionContext.Param, new RpcAuthorization(executionContext));
        }

        /// <summary>
        /// 调用Rpc服务
        /// </summary>
        /// <param name="executionContext">可以传null，但Rpc调用时，则无Identity和Token信息</param>
        /// <param name="serviceName">Rpc服务名称</param>
        /// <param name="apiName">Rpc Api名称</param>
        /// <param name="param">启用参数</param>
        public async Task<RpcResponse> InvokeRpcAsync(ExecutionContext executionContext, string serviceName, string apiName, object param)
        {
            return await InvokeAsync(serviceName, apiName, param, executionContext == null ? null : new RpcAuthorization(executionContext));
        }

        /// <summary>
        /// 调用Rpc服务。
        /// </summary>
        /// <param name="serviceName">Rpc服务名称</param>
        /// <param name="apiName">Rpc Api名称</param>
        /// <param name="param">启用参数</param>
        /// <param name="authInfo">用户授权信息</param>
        protected virtual async Task<RpcResponse> InvokeAsync(string serviceName, string apiName, object param, RpcAuthorization authorization)
        {
            var channel = await _grpcChannelProvider.Provide(serviceName);

            try
            {
                return await Policy.Handle<RpcException>()
                    .WaitAndRetry(3, i => TimeSpan.FromMilliseconds(100), async (e, s) => channel = await _grpcChannelProvider.Provide(serviceName))
                    .Execute(async() =>
                    {
                        var client = new RpcService.RpcServiceClient(channel.GetOrOpen());
                        var content = (param is JToken) ? (param as JToken).ToString(Formatting.None) : JsonConvert.SerializeObject(param);
                        var request = new RpcRequest
                        {
                            ServiceName = serviceName,
                            ApiName = apiName,
                            Params = content,
                            Identity = authorization?.Identity,
                            Token = authorization?.Token,
                            Client = clientName.Value
                        };

                        return await client.InvokeAsync(request);
                    });
            }
            catch (RpcException ex)
            {
                _logger.LogError(ex, $"Rpc invocation failed： `{serviceName}`.`{apiName}` in `{channel.Config.ServerAddress}`.");
                throw ex;
            }
        }

        private static Lazy<string> clientName = new Lazy<string>(GetClientName);

        private static string GetClientName()
        {
            var baseType = typeof(RpcClient);
            var stackFrames = new StackTrace().GetFrames();

            // 找到最近的一个从RpcRepository类继承的类，正常情况它会是调用的入口
            foreach (var stackFrame in stackFrames)
            {
                if (baseType != stackFrame.GetMethod().DeclaringType
                    && baseType.IsAssignableFrom(stackFrame.GetMethod().DeclaringType))
                {
                    return stackFrame.GetMethod().DeclaringType.Assembly.GetName().Name;
                }
            }

            var exclusiveNamePrefix = new string[] { "Wicture.DbRESTFul", "System", "Polly" };
            foreach (var stackFrame in stackFrames)
            {
                var name = stackFrame.GetMethod().DeclaringType.Assembly.GetName().Name;
                if (!exclusiveNamePrefix.Any(n => name.StartsWith(n, StringComparison.InvariantCultureIgnoreCase)))
                {
                    return name;
                }
            }

            return string.Empty;
        }
    }
}
